"""
Utility functions useful for Business Analyst - the glue functions not fitting neatly anywhere else.
"""
import asyncio
from functools import wraps, lru_cache
import importlib
from itertools import product
import re
import threading
from typing import Any, AnyStr, Iterable, Optional, Tuple, Union

from arcgis.features import FeatureSet
from arcgis.gis import GIS, User
from arcgis.geometry import Geometry, SpatialReference
import numpy as np
import pandas as pd


def module_avail(module_name: AnyStr) -> bool:
    """
    Determine if module is available in this environment.
    """
    if importlib.util.find_spec(module_name) is not None:
        avail = True
    else:
        avail = False
    return avail


# flag a couple of useful module availabilities
avail_arcpy = module_avail("arcpy")
avail_shapely = module_avail("shapely")


def local_vs_gis(fn):
    """Decorator to facilitate bridging between local and remote resources."""
    # get the method - this will be used to redirect the function call
    fn_name = fn.__name__

    @wraps(fn)
    def wrapped(self, *args, **kwargs):
        # get source no matter how defined
        for val in ["source", "_source", "_gis", "gis"]:
            if val in self.__dict__.keys():
                src = getattr(self, val)
                break

        # create the prefix string with the ablity to handle leading underscores...enabling underscored preprocessing
        prefix = fn_name if fn_name.startswith("_") else f"_{fn_name}"

        # if performing analysis locally, try to access the function locally, but if not implemented, catch the error
        if src == "local":
            try:
                fn_to_call = getattr(self, f"{prefix}_local")
            except AttributeError:
                raise NotImplementedError(
                    f"'{fn_name}' not available using 'local' as the source."
                )

        # now, if performing analysis using a Web GIS, then access the function referencing remote resources
        elif isinstance(src, GIS):
            try:
                fn_to_call = getattr(self, f"{prefix}_gis")
            except AttributeError:
                raise NotImplementedError(
                    f"'{fn_name}' not available using a Web GIS as the source."
                )

        # if another source, we don't plan on doing that any time soon
        else:
            raise NotImplementedError(f"'{self.source}' is not a recognized source.")

        return fn_to_call(*args, **kwargs)

    return wrapped


@lru_cache(maxsize=10)
def local_business_analyst_avail() -> bool:
    """
    Check if a local installation of Business Analyst is available.
    """
    avail = False

    if avail_arcpy is True:
        import arcpy

        if arcpy.CheckExtension("Business"):
            avail = True

    return avail


@lru_cache(maxsize=10)
def local_ba_data_avail() -> bool:
    """
    Check to see if any local business analyst data packs are installed.
    """
    avail = False

    if local_business_analyst_avail():
        # lazy load to avoid import issues
        import arcpy._ba

        # TODO: remove once bapy patched
        # addresses issue with bapy not being part of Notebook server docker image
        if "getLocalDatasets" in arcpy._ba.__dict__.keys() or module_avail("bapy"):
            # if data is available, there will be more than one dataset available
            avail = len(list(arcpy._ba.ListDatasets())) > 0

    return avail


def set_source(in_source: Optional[Union[str, GIS]] = None) -> Union[str, GIS]:
    """
    Helper function to check source input. The source can be set explicitly, but if nothing is provided, it
    assumes the order of local first and then a Web GIS. Along the way, it also checks to see if a GIS object
    instance is available in the current session.
    """
    # lazy import to load the active_gis
    from arcgis.env import active_gis

    # if string input is provided, should be 'local'
    if isinstance(in_source, str):
        # cast to lowercast
        in_source = in_source.lower()

        # account for setting to "pro"
        if in_source == "pro":
            in_source = "local"

        if in_source != "local":
            raise Exception(
                f'Source must be either "local" or a Web GIS instance, not {in_source}.'
            )

        elif in_source == "local" and not local_business_analyst_avail():
            raise Exception(
                f"If using local source, the Business Analyst extension must be available."
            )

        elif in_source == "local":
            source = "local"

    # if nothing provided, default to local if arcpy is available, and remote if arcpy not available
    elif in_source is None and local_business_analyst_avail() and local_ba_data_avail():
        source = "local"

    # if a web gis is not provided, check to see if working session has one available
    elif in_source is None and active_gis:
        source = active_gis

    # if not using local, use a GIS
    elif isinstance(in_source, GIS):
        source = in_source

    # provide a fall-though catch
    else:
        raise Exception(
            'Source must be either "local" if using ArcGIS Pro with Business Analyst, or a GIS object '
            "instance connected ot ArcGIS Enterprise with Business Analyst Server or ArcGIS Online."
        )

    return source


def _assert_privileges_access(user: User) -> None:
    """Helper function determining if can access user privileges."""
    assert "privileges" in user.keys(), (
        f"Cannot access privileges of {user.username}. Please ensure either this is "
        "your username, or you are logged in as an administrator to be able to "
        "view privileges."
    )


def get_helper_service_url(gis: GIS, service_key: str) -> str:
    """Helper function to ensure helper service is defined in gis properties."""
    err_msg = (
        f"It appears {service_key} is not available in your Web GIS. Please ensure a {service_key} service is "
        "defined."
    )
    assert service_key in gis.properties.helperServices, err_msg
    url = gis.properties.helperServices[service_key]["url"]
    assert len(url), err_msg
    return url


def can_enrich_gis(user: User) -> bool:
    """Determine if the provided user has data enrichment privileges in the Web GIS.

    .. note::

        The current user can be retrieved using `gis.users.me` for input.

    Args:
        user: Required `arcgis.gis.User` object instance.

    Returns:
       Boolean indicating if the provided user has enrichment privileges in the Web GIS.
    """
    # privileges may no be available
    _assert_privileges_access(user)

    # use list comprehension to account for future modifications to privilege descriptions
    bool_enrich = len([priv for priv in user["privileges"] if "enrich" in priv]) > 0

    return bool_enrich


def has_networkanalysis_gis(
    user: User,
    network_function: Optional[str] = None,
) -> bool:
    """Determine if the provided user has network analysis privileges in the Web GIS.

    .. note::

        The current user can be retrieved using `gis.users.me` for input.

    Args:
        user: Required `arcgis.gis.User` object instance.
        network_function: Optional string describing specific network function to check for.
            Valid values include 'closestfacility', 'locationallocation', 'optimizedrouting',
            'origindestinationcostmatrix', 'routing', 'servicearea', or 'vehiclerouting'.

    Returns: Boolean indicating if the provided user has network analysis privileges in the Web GIS.
    """
    # validate network_function if provided
    if network_function is not None:
        ntwrk_fn_lst = [
            "closestfacility",
            "locationallocation",
            "optimizedrouting",
            "origindestinationcostmatrix",
            "routing",
            "servicearea",
            "vehiclerouting",
        ]
        assert network_function in ntwrk_fn_lst, (
            f'The network function provided, f"{network_function}," is not in '
            f'the list of network functions [{", ".join(ntwrk_fn_lst)}.'
        )

    # privileges may not be available
    _assert_privileges_access(user)

    # get the network analysis capabilities from the privileges
    prv_tail_lst = [
        prv.split(":")[-1] for prv in user["privileges"] if "networkanalysis" in prv
    ]

    # if nothing was found, there are not any capabilities
    if len(prv_tail_lst) == 0:
        bool_net = False

    # if no specific network capability is being requested, check for overall access
    elif not network_function:
        bool_net = "networkanalysis" in prv_tail_lst

    # if a specific network function is being checked
    else:
        # just in case, ensure the input is lowercase
        network_function = network_function.lower()

        # check to ensure function is in capabilities determining if user has privileges
        bool_net = network_function in prv_tail_lst

    return bool_net


def geography_iterable_to_arcpy_geometry_list(
    geography_iterable: Union[pd.DataFrame, Iterable, Geometry],
    geometry_filter: Optional[str] = None,
) -> list:
    """
    Processing helper to convert a iterable of geography_levels to a list of ArcPy Geometry objects suitable for input
    into ArcGIS ArcPy geoprocessing tools.

    Args:
        geography_iterable:
            Iterable containing valid geographies.
        geometry_filter:
            String point|polyline|polygon to use for validation to ensure correct geometry type.

    Returns:
        List of ArcPy objects.
    """
    if not avail_arcpy:
        raise Exception(
            "Converting to ArcPy geometry requires an environment with ArcPy available."
        )

    # do some error checking on the geometry filter
    geometry_filter = (
        geometry_filter.lower() if geometry_filter is not None else geometry_filter
    )
    if (
        geometry_filter not in ["point", "polyline", "polygon"]
        and geometry_filter is not None
    ):
        raise Exception(
            f"geometry_filter must be point|polyline|polygon {geometry_filter}"
        )

    # if a DataFrame, check to ensure is spatial, and convert to list of arcgis Geometry objects
    if isinstance(geography_iterable, pd.DataFrame):
        if geography_iterable.spatial.validate() is True:
            geom_col = [
                col
                for col in geography_iterable.columns
                if geography_iterable[col].dtype.name.lower() == "geometry"
            ][0]
            geom_lst = list(geography_iterable[geom_col].values)
        else:
            raise Exception(
                "The provided geography_iterable DataFrame does not appear to be a Spatially Enabled "
                "DataFrame or if so, all geographies do not appear to be valid."
            )

    # accommodate handling pd.Series input
    elif isinstance(geography_iterable, pd.Series):
        if (
            "SHAPE" not in geography_iterable.keys()
            or geography_iterable.name != "SHAPE"
        ):
            raise Exception(
                "SHAPE geometry field must be in the pd.Series or the pd.Series must be the SHAPE to use "
                "a pd.Series as input."
            )

        # if just a single row passed in
        elif "SHAPE" in geography_iterable.keys():
            geom_lst = [geography_iterable["SHAPE"]]

        # otherwise, is a series of geometry objects, and just pass over to geom_lst since will be handled as iterable
        else:
            geom_lst = geography_iterable

    # if a list, ensure all child objects are polygon geographies and convert to list of arcpy.Geometry objects
    elif isinstance(geography_iterable, list):
        for geom in geography_iterable:
            if not isinstance(geom, Geometry):
                raise Exception(
                    "The provided geographies in the selecting_geometry list do not appear to all be "
                    "valid."
                )
        geom_lst = geography_iterable

    # if a single geometry object instance, ensure is polygon and make into single item list of arcpy.Geometry
    elif isinstance(geography_iterable, Geometry):
        geom_lst = [geography_iterable]

    else:
        raise Exception(
            "geography_iterable must be either a Spatially Enabled Dataframe, pd.Series with a SHAPE "
            f"column, list or single geometry object - not {type(geography_iterable)}."
        )

    # ensure all geographies are correct geometry type if provided
    if geometry_filter is not None:
        for geom in geom_lst:
            if geom.geometry_type != geometry_filter:
                raise Exception(
                    "geography_iterable geographies must be polygons. It appears you have provided at "
                    f'least one "{geom.geometry_type}" geometry.'
                )

    # convert the objects in the geometry list to ArcPy Geometry objects
    arcpy_lst = [geom.as_arcpy for geom in geom_lst]

    return arcpy_lst


def get_sanitized_names(names: Union[str, list, tuple, pd.Series]) -> pd.Series:
    """
    Sanitize the column names to be PEP 8 compliant.
    Useful when trying to match up column names from previously enriched data.

    Args:
        names: Iterable (list, tuple, pd.Series) of names to be "sanitized"

    Returns:
        pd.Series of column names
    """
    # if just one name was passed in, make it an iterable
    names = [names] if isinstance(names, str) else names

    # sanititze the column names and drop into a series
    sani_names = pd.Series(map(pep8ify, names))

    return sani_names


def validate_spatial_reference(
    spatial_reference: Union[str, int, dict, SpatialReference]
) -> SpatialReference:
    """Validate the variety of ways a spatial reference can be inputted. This does not validate the WKID."""
    # instantiate the output spatial reference variable
    sr = None

    if isinstance(spatial_reference, dict):
        assert "wkid" in spatial_reference.keys(), (
            "If providing spatial reference as a dictionary, it must conform "
            'look like {"wkid": <WKID>}, such as {"wkid": 4326}.'
        )
        sr = spatial_reference
    elif isinstance(spatial_reference, str):
        assert spatial_reference.isdecimal(), (
            "If providing a string to identify a spatial reference, it must be a "
            "WKID for your desired spatial reference."
        )
        sr = {"wkid": spatial_reference}
    elif isinstance(spatial_reference, int):
        sr = {"wkid": spatial_reference}
    elif isinstance(spatial_reference, SpatialReference):
        sr = SpatialReference
    elif avail_arcpy:
        import arcpy

        if isinstance(spatial_reference, arcpy.SpatialReference):
            sr = SpatialReference(spatial_reference.factoryCode)

    # if nothing has been found, something is wrong
    if sr is None:
        raise Exception(
            "The spatial reference must be either a string or integer specify the WKID, a dictionary "
            'specifying the WKID such as {"wkid": 4326}, or a SpatialReference object.'
        )

    return sr


def is_dict_geometry(in_dict: dict) -> bool:
    """Determine if input dictionary is a Geometry object."""
    if (
        ("x" in in_dict.keys() and "y" in in_dict.keys())
        or ("points" in in_dict.keys())
        or ("ringCurves" in in_dict.keys())
        or ("rings" in in_dict.keys())
        or ("paths" in in_dict.keys())
        or ("pathCurves" in in_dict.keys())
    ):
        is_geometry = True
    else:
        is_geometry = False
    return is_geometry


def is_dict_featureset(in_dict: dict) -> bool:
    """Determine if input dictionary is a FeatureSet."""
    if isinstance(in_dict, dict):
        is_featureset = "features" in in_dict.keys() and "fields" in in_dict.keys()
    else:
        is_featureset = False
    return is_featureset


def get_spatially_enabled_dataframe(
    input_object: Union[
        pd.DataFrame, pd.Series, Geometry, FeatureSet, Iterable, np.ndarray
    ],
    spatial_column: str = "SHAPE",
) -> pd.DataFrame:
    """Garbage disposal taking variety of possible inputs and outputting, if possible, a Pandas Spatially Enabled
    DataFrame."""
    # ensure only one FeatureSet getting passed in if an iterable is passed in
    if isinstance(input_object, Iterable) and not isinstance(
        input_object, pd.DataFrame
    ):
        if is_dict_featureset(input_object[0]) or isinstance(input_object, FeatureSet):
            assert len(input_object) == 1, "Only one FeatureSet can be used for input"

            # pop out the FeatureSet if this is passed in
            if isinstance(input_object[0], FeatureSet):
                input_object = input_object[0]

    # check if is FeatureSet dict and convert to FeatureSet object if it is
    if is_dict_featureset(input_object):
        input_object = FeatureSet(input_object)

    # if a FeatureSet, convert to spatially enabled data frame
    if isinstance(input_object, FeatureSet):
        input_object = input_object.sdf

    # if just a single geometry passed in, we need to get it into a list
    if isinstance(input_object, (Geometry, dict)):
        input_object = [input_object]

    # if any type of iterable other than a series, make into a series
    if isinstance(input_object, (Iterable, np.ndarray)) and not isinstance(
        input_object, pd.DataFrame
    ):
        # Geometry objects may be passed in as an iterable of dicts - convert to Geometry if this is the case
        first_obj = input_object[0]
        if isinstance(first_obj, dict):
            if is_dict_geometry(first_obj):
                input_object = [Geometry(obj) for obj in input_object]

        input_object = pd.Series(input_object)

    # at this juncture, the only real options are either a Series or DataFrame, so if Series, make into DataFrame
    if isinstance(input_object, pd.Series):
        input_object = input_object.to_frame("SHAPE")

    # if the geometry has not been set, take care of it
    if input_object.spatial._name is None:
        assert spatial_column in input_object.columns, (
            f"The spatial column cannot be set to {spatial_column}, "
            f"because it is not a column in the input DataFrame."
        )
        input_object.spatial.set_geometry(spatial_column)

    return input_object


def preproces_code_inputs(codes):
    """helper function to preprocess naics or sic codes"""
    if isinstance(codes, (str, int)):
        codes = [codes]

    elif isinstance(codes, (pd.Series, list, tuple, np.ndarray)):
        codes = [str(cd) if isinstance(cd, int) else cd for cd in codes]

    return codes


def get_top_codes(codes: Union[pd.Series, list, tuple], threshold: float = 0.5) -> list:
    """Get the top category codes by only keeping those compromising 50% or greater of the records.

    Args:
        codes: Iterable, preferable a Pandas Series, of code values to filter.
        threshold: Decimal value representing the proportion of values to use for creating
            the list of top values.

    Returns:
        List of unique code values.
    """
    # check the threshold to ensure it is deicmal
    msg_thrshld = (
        f'"threshold" must be a decimal value between zero and one, not {threshold}'
    )
    assert 0 < threshold < 1, msg_thrshld

    # ensure the input codes iterable is a Pandas Series
    cd_srs = codes if isinstance(codes, pd.Series) else pd.Series(codes)

    # get the instance count for each unique code value
    cnt_df = cd_srs.value_counts().to_frame("cnt")

    # calculate the percent each of the codes comprises of the total values as a decimal
    cnt_df["pct"] = cnt_df.cnt.apply(lambda x: x / cnt_df.cnt.sum())

    # calculate a running total of the percent for each value (running total percent)
    cnt_df["pct_cumsum"] = cnt_df["pct"].cumsum()

    # finally, get the values comprising the code values for the top threshold
    cd_vals = list(
        cnt_df[(cnt_df["pct_cumsum"] < threshold) | (cnt_df["pct"] > threshold)].index
    )

    return cd_vals


def pep8ify(name):
    """PEP8ify name"""
    if name is None:
        res = None
    else:
        if "." in name:
            name = name[name.rfind(".") + 1 :]
        if name[0].isdigit():
            name = "level_" + name
        name = name.replace(".", "_")
        if "_" in name:
            name = name.lower()
        s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        s2 = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()
        res = s2.replace(" ", "")
    return res


def pro_at_least_version(version: str) -> bool:
    """
    Test the current ArcGIS Pro version to be equal or greater than.
    Args:
        version: Version number eg ('2', '2.1', '2.8.1')
    Returns:
        Boolean indicating if version is at least or greater than specified version.
    """

    # late import since may or may not already be loaded
    import arcpy

    # get parts of the current version
    v_lst = [int(v) for v in arcpy.GetInstallInfo()["Version"].split(".")]

    # get parts of the version to test for
    in_lst = [int(v) for v in version.split(".")]

    # extend the shorter version list to match the longer
    max_len = max((len(v_lst), len(in_lst)))
    v_lst, in_lst = [lst + ([0] * (max_len - len(lst))) for lst in [v_lst, in_lst]]

    # variable to store status
    at_least = False
    all_parts_equal = True  # until proven false

    # test all the parts of the input version against the current version
    for idx in range(0, max_len):
        # evaluate if the part and if greater, break and report status
        if v_lst[idx] > in_lst[idx]:
            all_parts_equal = False
            at_least = True  # current Pro version is more recent
            break

        if v_lst[idx] < in_lst[idx]:
            all_parts_equal = False
            at_least = False  # current Pro version is too old
            break

    return (
        at_least | all_parts_equal
    )  # if versions are equal, return true. Otherwise return at_least


def extract_from_kwargs(paramater_key: str, kwargs: dict) -> Tuple[Any, dict]:
    """
    Provide ability to unpack value from kwargs in relatively streamlined fashion.
    Args:
        paramater_key: String key of value to be extracted from kwargs.
        kwargs: The dictionary of keyword arguments for the value to be extracted from.
    Returns:
        Tuple of the parameter value proided as input and the kwargs updated to no longer
        include the parameter retrieved.
    """
    if paramater_key in kwargs.keys():
        param_val = kwargs[paramater_key]
        del kwargs[paramater_key]
    else:
        param_val = None
    return param_val, kwargs


def validate_network_travel_mode(
    source, travel_mode: str, proximity_metric: Optional[str]
):
    """Validate the travel_mode string or index."""
    # dictionary of potential aliases for travel modes
    travel_mode_dict = {"walk": "walking", "drive": "driving", "truck": "trucking"}

    # flag for ensuring a travel mode match is found
    travel_mode_match = False

    # if the travel mode was provided as a string, get the correct travel mode alias
    if isinstance(travel_mode, str):
        # switch to all lowercase to mitigate case discrepancies
        travel_mode = travel_mode.lower()

        # if generic given, change to correct generic [Take care of old code for Buffer Study Area]
        if travel_mode in ["driving", "walking", "trucking"]:
            travel_mode_type = "time"
            if proximity_metric:
                proximity_metric = proximity_metric.lower()
                if proximity_metric not in ["seconds", "minutes", "hours", "days"]:
                    travel_mode_type = "distance"
            travel_mode = "{} {}".format(travel_mode, travel_mode_type)

        # if the proximity type is straight line, just make sure in correct format (used for enrich method)
        if travel_mode == "straight_line" or travel_mode == "Straight Line":
            travel_mode = "Straight Line"
            travel_mode_match = True

        # tru to find a transportation travel mode
        else:
            # account for potential differences in descriptions from alias dict by building a list of candidates
            prx_lst = [travel_mode]

            for key, val in travel_mode_dict.items():
                if key in travel_mode and val not in travel_mode:
                    prx_lst.append(travel_mode.replace(key, val))

            # look in the name and alias columns for a match
            for col, prx_mthd in product(["name", "alias"], prx_lst):
                # try to find a match in the column for the proximity method
                tmp_df = source.travel_modes[
                    source.travel_modes[col].str.lower() == prx_mthd
                ]

                # if a match is found, pluck out the alias
                if len(tmp_df):
                    travel_mode = tmp_df.iloc[0]["alias"]
                    travel_mode_match = True
                    break

    # check if potential match by travel mode index
    elif isinstance(travel_mode, int):
        assert travel_mode < len(source.travel_modes.index)
        travel_mode = source.travel_modes.iloc[travel_mode]["alias"]
        travel_mode_match = True

    # ensure a recognized proximity type was found
    assert (
        travel_mode_match
    ), f"The travel mode provided, {travel_mode}, is not recognized as an available travel mode."

    return travel_mode


def add_proximity_to_enrich_feature(
    source: GIS,
    feature: dict,
    travel_mode: str = "straight_line",
    proximity_metric: Optional[str] = None,
    proximity_value: int = 1,
    proximity_area_overlap: bool = True,
) -> dict:
    """Add proximity metrics onto a feature in a feature set for sending to the enrich REST endpoint."""
    # alias list to standardize the proximity_metric input
    if proximity_metric is not None:
        trvl_md_aliases = {
            "kilometers": ["kilometer", "km"],
            "miles": ["mile"],
            "minutes": ["min"],
        }

        # ensure lowercase to avoid issues with case
        proximity_metric = proximity_metric.lower()

        # look through the names and aliases to see if we can find one
        for val, alias_lst in trvl_md_aliases.items():
            if proximity_metric == val or proximity_metric in alias_lst:
                proximity_metric = val
                break

    # if just doing a buffer, set the correct area type and set variable for travel mode type
    if travel_mode is None:
        travel_mode = "straight_line"
    if travel_mode == "straight_line":
        feature["areaType"] = (
            "RingBuffer" if proximity_area_overlap else "RingBufferBands"
        )
        trvl_md_typ = "distance"

    # otherwise, doing a network type and need to figure out what the travel mode is
    else:
        feature["areaType"] = "NetworkServiceArea"

        # scrub the travel mode
        travel_mode = validate_network_travel_mode(
            source, travel_mode, proximity_metric
        )

        # pull out the category from the travel modes and set the travel mode flat (temporal or distance)
        source_travel_mode = source.travel_modes[
            source.travel_modes["alias"] == travel_mode
        ]
        trvl_md_typ = source_travel_mode.iloc[0]["impedance_category"]

        feature["travel_mode"] = source_travel_mode.iloc[0]["travel_mode_dict"]

        # tack on polygon area overlap
        if proximity_area_overlap:
            feature["networkOptions"] = {"polygon_overlap_type": "Disks"}
        else:
            feature["networkOptions"] = {"polygon_overlap_type": "Rings"}

    # if no proximity metric provided, provide default based on travel mode, and also validate if provided
    if proximity_metric is None:
        if trvl_md_typ == "distance":
            proximity_metric = "kilometers"
        elif trvl_md_typ == "temporal":
            proximity_metric = "minutes"

    # set the buffer units if now populated
    if proximity_metric is not None:
        feature["bufferUnits"] = proximity_metric

    # if the proximity_value is a single value
    if isinstance(proximity_value, (int, float)):
        proximity_value = [proximity_value]

    # if some other iterable was used for input, make sure a list
    if not isinstance(proximity_value, list):
        proximity_value = proximity_value.split(",")

    # put the scalar proximity value(s) in the payload
    feature["bufferRadii"] = proximity_value

    return feature


def add_proximity_to_enrich_feature_list(
    source: GIS,
    feature_list: Iterable,
    travel_mode: str = "straight_line",
    proximity_metric: Optional[str] = None,
    proximity_value: int = 1,
    proximity_area_overlap: bool = True,
) -> list:
    """Add proxmity metrics to a FeatureSet for sending to the enrich REST endpoint."""
    prx_feat_lst = [
        add_proximity_to_enrich_feature(
            source,
            f,
            travel_mode,
            proximity_metric,
            proximity_value,
            proximity_area_overlap,
        )
        for f in feature_list
    ]
    return prx_feat_lst


# asynchronous function support section - solution located on StackOverflow
# https://stackoverflow.com/questions/55409641/asyncio-run-cannot-be-called-from-a-running-event-loop
class RunThread(threading.Thread):
    def __init__(self, func, args, kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.result = None
        super().__init__()

    def run(self):
        self.result = asyncio.run(self.func(*self.args, **self.kwargs))


def run_async(func, *args, **kwargs):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None
    if loop and loop.is_running():
        thread = RunThread(func, args, kwargs)
        thread.start()
        thread.join()
        return thread.result
    else:
        return asyncio.run(func(*args, **kwargs))



from arcgis.gis.sharing._users import UserInvitationManager
from arcgis.gis.sharing._dependency import DependencyManager
from arcgis.gis.sharing._marketplace import MarketPlaceManager


__all__ = ["UserInvitationManager", "DependencyManager", "MarketPlaceManager"]



"""
Entry point to working with local enterprise GIS functions
"""
from datetime import datetime
from typing import Optional
from ...gis._impl._con import Connection
from ...gis import GIS, Item, User
from ._resources import PortalResourceManager
from ._base import BasePortalAdmin
from ...apps.tracker._location_tracking import LocationTrackingManager


########################################################################
class PortalAdminManager(BasePortalAdmin):
    """
    This is the root resource for administering your portal. Starting from
    this root, all of the portal's environment is organized into a
    hierarchy of resources and operations. A version number is returned as
    a part of this resource. After installation, the portal can be
    configured using the Create Site operation. Once initialized, the
    portal environment is available through System and Security resources.

    ================    =================================================================================
    **Parameter**        **Description**
    ----------------    ---------------------------------------------------------------------------------
    url                 web address to portaladmin rest API (ends with: portal//sharing/rest/)
    ----------------    ---------------------------------------------------------------------------------
    gis                 GIS object containing Administrative credentials
    ----------------    ---------------------------------------------------------------------------------
    initialize          Optional if True, properties of REST endpoint are loaded on creation of object.
                        False (default) means they are loaded when needed.
    ================    =================================================================================

    """

    _logs = None
    _federation = None
    _system = None
    _security = None
    _machines = None
    _site = None
    _url = None
    _gis = None
    _idp = None
    _ux = None
    _sp = None
    _metadata = None
    _collaborations = None
    _servers = None
    _pp = None
    _license = None
    _livingatlas = None
    _category_schema = None
    _whm = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis=None, **kwargs):
        """initializer"""
        if kwargs.pop("is_admin", True):
            super(PortalAdminManager, self).__init__(url=url, gis=gis, **kwargs)
            initialize = kwargs.pop("initialize", False)
            if isinstance(gis, Connection):
                self._con = gis
            elif isinstance(gis, GIS):
                self._gis = gis
                self._con = gis._con
            else:
                raise ValueError("connection must be of type GIS or Connection")
            try:
                self.resources = PortalResourceManager(gis=self._gis)
            except:
                pass
            if initialize:
                self._init(self._gis)
        else:
            super(PortalAdminManager, self).__init__(
                url=url, gis=gis, is_admin=False, initialize=False
            )
            if isinstance(gis, Connection):
                self._con = gis
            elif isinstance(gis, GIS):
                self._gis = gis
                self._con = gis._con

    # ----------------------------------------------------------------------
    @property
    def ux(self):
        """
        returns a UX/UI manager with properties such as description, featured_content, name, etc.

        :return:
            :class:`~arcgis.gis.admin.UX` object

        """
        if self._ux is None:
            from ._ux import UX

            self._ux = UX(gis=self._gis)
        return self._ux

    # ----------------------------------------------------------------------
    @property
    def collaborations(self):
        """
        The collaborations resource lists all collaborations in which a
        portal participates

        :return:
            :class:`~arcgis.gis.admin.CollaborationManager` object

        """
        if self._collaborations is None:
            from ._collaboration import CollaborationManager

            self._collaborations = CollaborationManager(gis=self._gis)
        return self._collaborations

    # ----------------------------------------------------------------------
    @property
    def category_schema(self):
        """
        This resource allows for the setting and manipulating of category schemas.

        :return:
            :class:`~arcgis.gis.admin.CategoryManager` object

        """
        if self._category_schema is None:
            from ._catagoryschema import CategoryManager

            self._category_schema = CategoryManager(gis=self._gis)
        return self._category_schema

    # ----------------------------------------------------------------------
    @property
    def idp(self):
        """
        This resource allows for the setting and configuration of the identity provider

        :return:
            :class:`~arcgis.gis.admin.IdentityProviderManager` object

        """
        if self._idp is None:
            from ._idp import IdentityProviderManager

            self._idp = IdentityProviderManager(gis=self._gis)
        return self._idp

    # ----------------------------------------------------------------------
    @property
    def location_tracking(self):
        """

        The manager for Location Tracking. See :class:`~arcgis.apps.tracker.LocationTrackingManager`.

        :return:
            :class:`~arcgis.apps.tracker.LocationTrackingManager` object

        """
        return LocationTrackingManager(self._gis)

    # ----------------------------------------------------------------------
    @property
    def social_providers(self):
        """
        This resource allows for the setting and configuration of the social providers
        for a GIS.

        :return:
            :class:`~arcgis.gis.admin.SocialProviders` object

        """
        if self._sp is None:
            from ._socialproviders import SocialProviders

            self._sp = SocialProviders(gis=self._gis)
        return self._sp

    @property
    def info(self) -> dict:
        """
        Returns the current version and build number of the Enterprise system

        :returns: dict
        """
        if self._gis.version >= [10, 3]:
            url = "%s/portaladmin/info" % self._gis._portal.url
            params = {"f": "json"}
            return self._gis._con.get(url, params)
        return None

    # ----------------------------------------------------------------------
    @property
    def metadata(self):
        """
        returns a set of tools to work with ArcGIS Enterprise metadata
        settings.

        :return:
            :class:`~arcgis.gis.admin.MetadataManager` object

        """
        if self._metadata is None:
            from ._metadata import MetadataManager

            self._metadata = MetadataManager(gis=self._gis)
        return self._metadata

    # ----------------------------------------------------------------------
    @property
    def servers(self):
        """returns a server manager object

        :return:
            :class:`~arcgis.gis.server.ServerManager`

        """
        if self._servers is None:
            from ..server import ServerManager

            self._servers = ServerManager(gis=self._gis)
        return self._servers

    # ----------------------------------------------------------------------
    def scheduled_tasks(
        self,
        item: Optional[Item] = None,
        active: Optional[bool] = None,
        user: Optional[User] = None,
        types: Optional[str] = None,
    ):
        """
        This property allows `org_admins` to be able to see all scheduled tasks on the enterprise

        ================  ===============================================================================
        **Parameter**      **Description**
        ----------------  -------------------------------------------------------------------------------
        item              Optional Item. The item to query tasks about.
        ----------------  -------------------------------------------------------------------------------
        active            Optional Bool. Queries tasks based on active status.
        ----------------  -------------------------------------------------------------------------------
        user              Optional User. Search for tasks for a single user.
        ----------------  -------------------------------------------------------------------------------
        types             Optional String. The type of notebook execution for the item.  This can be
                          `ExecuteNotebook`, or `UpdateInsightsWorkbook`.
        ================  ===============================================================================


        :return: List of Tasks

        """
        _tasks = []
        num = 100
        url = f"{self._gis._portal.resturl}portals/self/allScheduledTasks"
        params = {"f": "json", "start": 1, "num": num}
        if item:
            params["itemId"] = item.itemid
        if not active is None:
            params["active"] = active
        if user:
            params["userFilter"] = user.username
        if types:
            params["types"] = types
        res = self._con.get(url, params)
        start = res["nextStart"]
        _tasks.extend(res["tasks"])
        while start != -1:
            params["start"] = start
            params["num"] = num
            res = self._con.get(url, params)
            if len(res["tasks"]) == 0:
                break
            _tasks.extend(res["tasks"])
            start = res["nextStart"]
        return _tasks

    # ----------------------------------------------------------------------
    @property
    def machines(self):
        """
        This resource lists all the portal machines in a site. Each portal
        machine has a status that indicates whether the machine is ready
        to accept requests.


        :return:
            :class:`~arcgis.gis.admin.Machines` object

        """
        if self._machines is None:
            from ._machines import Machines

            # url root needs to be administrator site root
            url = "%s/portaladmin/machines" % self._gis._portal.url
            self._machines = Machines(url=url, gis=self._gis, portaladmin=self)
        return self._machines

    # ----------------------------------------------------------------------
    @property
    def security(self):
        """
        accesses the controls for the security of a local portal site

        :return:
            :class:`~arcgis.gis.admin.Security` object

        """

        if self._security is None:
            from ._security import Security

            url = "%s/portaladmin/security" % self._gis._portal.url
            self._security = Security(url=url, gis=self._gis)
        return self._security

    # ----------------------------------------------------------------------
    @property
    def site(self):
        """
        Site is the root resources used after a local GIS is installed. Here
        administrators can create, export, import, and join sites.

        :return:
            :class:`~arcgis.gis.admin.Site` object

        """
        if self._site is None:
            from ._site import Site

            self._site = Site(url=self._url, portaladmin=self)
        return self._site

    # ----------------------------------------------------------------------
    @property
    def logs(self):
        """
        returns a class to work with the portal logs

        :return:
            :class:`~arcgis.gis.admin.Logs` object

        """
        if self._logs is None:
            from ._logs import Logs

            url = "%s/portaladmin/logs" % self._gis._portal.url
            self._logs = Logs(url=url, gis=self._gis)
        return self._logs

    # ----------------------------------------------------------------------
    @property
    def federation(self):
        """
        provides access into the federation settings of a server.


        :return:
            :class:`~arcgis.gis.admin.Federation` object

        """
        if self._federation is None:
            from ._federation import Federation

            url = "%s/portaladmin/federation" % self._gis._portal.url
            self._federation = Federation(url=url, gis=self._gis)
        return self._federation

    # ----------------------------------------------------------------------
    @property
    def system(self):
        """
        This resource provides access to the ArcGIS Web Adaptor
        configuration, portal directories, database management server,
        indexing capabilities, license information, and the properties of
        your portal.


        :return:
            :class:`~arcgis.gis.admin.System` object

        """
        if self._system is None:
            from ._system import System

            url = "%s/portaladmin/system" % self._gis._portal.url
            self._system = System(url=url, gis=self._gis)
        return self._system

    # ----------------------------------------------------------------------
    @property
    def password_policy(self):
        """tools to manage a Site's password policy

        :return:
            :class:`~arcgis.gis.admin.PasswordPolicy` object

        """
        if self._pp is None:
            from ._security import PasswordPolicy

            url = "%s/portals/self/securityPolicy" % (self._gis._portal.resturl)
            self._pp = PasswordPolicy(url=url, gis=self._gis)
        return self._pp

    # ----------------------------------------------------------------------
    @property
    def license(self):
        """
        provides a set of tools to access and manage user licenses and
        entitlements.

        :return:
            :class:`~arcgis.gis.admin.LicenseManager` object

        """
        if self._license is None:
            from ._license import LicenseManager

            url = f"{self._gis._portal.resturl}/portals/self/purchases"
            self._license = LicenseManager(url=url, gis=self._gis)
        return self._license

    # ----------------------------------------------------------------------
    @property
    def living_atlas(self):
        """
        provides a set of tools to manage and setup Living Atlas content.

        :return:
            :class:`~arcgis.gis.admin.LivingAtlas` object

        """
        if self._livingatlas is None:
            from ._livingatlas import LivingAtlas

            url = self._url + "/system/content/livingatlas"
            self._livingatlas = LivingAtlas(url=url, gis=self._gis)
        return self._livingatlas

    # ----------------------------------------------------------------------
    @property
    def webhooks(self):
        """Provides access to Portal's WebHook Manager

        :return:
            :class:`~arcgis.gis.admin.WebhookManager` object

        """
        if self._whm is None and self._gis.version >= [6, 4]:
            from ._wh import WebhookManager

            url = self._gis._portal.resturl + "portals/self/webhooks"
            self._whm = WebhookManager(url=url, gis=self._gis)
        return self._whm

    # ----------------------------------------------------------------------
    @property
    def mode(self) -> dict:
        """
        Gets/Set the mode of the ArcGIS Enterprise deployment.  When obtaining
        the mode, it returns information about the current state of the system.

        ================  ===============================================================================
        **Key**           **Description**
        ----------------  -------------------------------------------------------------------------------
        read_only         Required Boolean.  A boolean that specifies whether the Enterprise portal is
                          in read-only mode. Read-only mode will block requests to modify or create any
                          data, including content, users, groups, or site settings.
                          The default value is false.
        ----------------  -------------------------------------------------------------------------------
        message           Optional String. Sets a custom message to be displayed whenever an attempt to
                          modify or update content or site settings is made through the API.
        ================  ===============================================================================

        .. code-block:: python

            >>> gis.admin.mode({'read_only' : False})
            >>> assert gis.admin.mode['isReadOnly'] == False

        """
        url = "%s/portaladmin/mode" % self._gis._portal.url
        params = {"f": "json"}
        return self._con.get(url, params)

    # ----------------------------------------------------------------------
    @mode.setter
    def mode(self, mode: dict):
        """
        See main ``mode`` property docstring.
        """
        url = "%s/portaladmin/mode/update" % self._gis._portal.url
        if mode is None:
            mode = {"read_only": False, "message": ""}
        params = {"f": "json", "isReadOnly": mode.pop("read_only", False)}
        if "message" in mode:
            params["description"] = mode.pop("message", "")
        res = self._con.post(url, params)
        if "status" in res and res["status"] != "success":
            raise RuntimeError(res)

    # ----------------------------------------------------------------------
    def history(
        self, start_date: datetime, num: int = 100, save_folder: Optional[str] = None
    ):
        """
        Returns a CSV file containing the login history from a start_date to the present.

        ================  ===============================================================================
        **Parameter**      **Description**
        ----------------  -------------------------------------------------------------------------------
        start_date        Required datetime.datetime object. The beginning date.
        ----------------  -------------------------------------------------------------------------------
        num               Optional Integer. The maximum number of records to return.
        ----------------  -------------------------------------------------------------------------------
        save_folder       Optional String. The save location of the CSV file.
        ================  ===============================================================================

        :return: string

        """
        if self._gis.properties.isPortal:
            raise NotImplementedError("Login history only available in ArcGIS Online")
        else:
            if self._gis.version >= [6.4]:
                import tempfile, json
                from arcgis._impl.common._utils import _date_handler

                if save_folder is None:
                    save_folder = tempfile.gettempdir()
                url = "{url}portals/self/history".format(url=self._gis._portal.resturl)
                params = {
                    "f": "csv",
                    "num": num,
                    "all": True,
                    "fromDate": json.dumps(start_date, default=_date_handler),
                }
                return self._gis._con.post(
                    url, params, file_name="history.csv", out_folder=save_folder
                )



from typing import ClassVar

from arcgis.realtime.velocity.feeds.mqtt import MQTT


class CiscoEdgeIntelligence(MQTT):
    """
    Receive events from a Cisco Edge Intelligence broker. This data class can be used to define the feed configuration
    and to create the feed.

    =====================       ========================================================================================
    **Parameter**                **Description**
    ---------------------       ----------------------------------------------------------------------------------------
    label                       String. Unique label for this feed instance.
    ---------------------       ----------------------------------------------------------------------------------------
    description                 String. Feed description.
    ---------------------       ----------------------------------------------------------------------------------------
    host                        String. Hostname of the Cisco Edge Intelligence broker prefixed with
                                "``tcp://``" for non-SSL or "``ssl://``" for SSL connections.
    ---------------------       ----------------------------------------------------------------------------------------
    port                        String. Port on which the Cisco Edge Intelligence broker is accessible.
    ---------------------       ----------------------------------------------------------------------------------------
    topic                       String. Topic over which event messages stream.
    ---------------------       ----------------------------------------------------------------------------------------
    qos_level                   int. Quality of Service (QoS) level defines the guarantee of delivery for a specific
                                message. In MQTT 3.1.1, a QoS of 0 means a message is delivered at most once, a QoS of 1
                                at least once, and a QoS of 2 exactly once. The default is: 0.
    =====================       ========================================================================================

    =====================       ========================================================================================
    **Optional Argument**       **Description**
    =====================       ========================================================================================
    username                    Username for basic authentication.
    ---------------------       ----------------------------------------------------------------------------------------
    password                    Password for basic authentication.
    ---------------------       ----------------------------------------------------------------------------------------
    client_id                   The client ID ArcGIS Velocity will use to connect to the Cisco Edge
                                Intelligence broker.
    ---------------------       ----------------------------------------------------------------------------------------
    data_format                 [:class:`~arcgis.realtime.velocity.input.EsriJsonFormat`, :class:`~arcgis.realtime.velocity.input.GeoJsonFormat`, :class:`~arcgis.realtime.velocity.input.DelimitedFormat`, :class:`~arcgis.realtime.velocity.input.JsonFormat`, :class:`~arcgis.realtime.velocity.input.XMLFormat`].
                                An instance that contains the data format configuration for this feed. Configure only
                                allowed formats. If this is not set right during initialization, a format will be
                                auto-detected and set from a sample of the incoming data. This sample will be fetched
                                from the configuration provided so far in the init.
    ---------------------       ----------------------------------------------------------------------------------------
    track_id_field              String. Name of the field from the incoming data that should be set as
                                track ID.
    ---------------------       ----------------------------------------------------------------------------------------
    geometry                    [:class:`~arcgis.realtime.velocity.feeds.XYZGeometry`, :class:`~arcgis.realtime.velocity.feeds.SingleFieldGeometry`]. An instance of geometry configuration
                                that will be used to create geometry objects from the incoming data.
    ---------------------       ----------------------------------------------------------------------------------------
    time                        [:class:`~arcgis.realtime.velocity.feeds.TimeInstant`, :class:`~arcgis.realtime.velocity.feeds.TimeInterval`]. An instance of time configuration that
                                will be used to create time information from the incoming data.
    =====================       ========================================================================================

    :return: A data class with cisco edge intelligence feed configuration.

    .. code-block:: python

        # Usage Example

        from arcgis.realtime.velocity.feeds import CiscoEdgeIntelligence
        from arcgis.realtime.velocity.feeds.geometry import XYZGeometry, SingleFieldGeometry
        from arcgis.realtime.velocity.feeds.time import TimeInterval, TimeInstant

        cisco_edge_config = CiscoEdgeIntelligence(
            label="feed_name",
            description="feed_description",
            host="cisco_host",
            port="cisco_port",
            topic="cisco_topic",
            qos_level=0,
            username="cisco_username",
            password="cisco_password",
            client_id="cisco_client_id",
            data_format=None
        )

        # use velocity object to get the FeedsManager instance
        feeds = velocity.feeds

        # use the FeedsManager object to create a feed from this feed configuration
        cisco_edge_feed = feeds.create(cisco_edge_config)
        cisco_edge_feed.start()
        feeds.items

    """

    # FeedTemplate properties
    _name: ClassVar[str] = "kinetic"



import traceback
from .._data import _raise_fastai_import_error
from ._inference_only_models import InferenceOnlyModel

HAS_TRANSFORMER = True

try:
    import torch
    from transformers import pipeline, logging
    from fastprogress.fastprogress import progress_bar

    try:
        # For version 3.3.0
        from transformers.modeling_auto import MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING
    except ModuleNotFoundError as e:
        # For version 4.5.1
        from transformers.models.auto.modeling_auto import (
            MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING,
        )
    EXPECTED_MODEL_TYPES = [
        x.__name__.replace("Config", "")
        for x in MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING.keys()
    ]
except Exception as e:
    transformer_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_TRANSFORMER = False
    EXPECTED_MODEL_TYPES = []


class ZeroShotClassifier(InferenceOnlyModel):
    """
    Creates a :class:`~arcgis.learn.text.ZeroShotClassifier` Object.
    Based on the Hugging Face transformers library

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    backbone                Optional string. Specifying the HuggingFace
                            transformer model name which will be used to
                            predict the answers from a given passage/context.

                            To learn more about the available models for
                            zero-shot-classification task, kindly visit:-
                            https://huggingface.co/models?pipeline_tag=zero-shot-classification
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    pretrained_path         Option str. Path to a directory, where pretrained
                            model files are saved.
                            If pretrained_path is provided, the model is
                            loaded from that path on the local disk.
    ---------------------   -------------------------------------------
    working_dir             Option str. Path to a directory on local filesystem.
                            If directory is not present, it will be created.
                            This directory is used as the location to save the
                            model.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.text.ZeroShotClassifier` Object
    """

    #: supported transformer architectures
    supported_backbones = EXPECTED_MODEL_TYPES

    def __init__(self, backbone=None, **kwargs):
        if not HAS_TRANSFORMER:
            _raise_fastai_import_error(import_exception=transformer_exception)
        super().__init__(backbone=backbone, task="zero-shot-classification", **kwargs)

    def predict(self, text_or_list, candidate_labels, show_progress=True, **kwargs):
        """
        Predicts the class label(s) for the input text

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required string or list. The sequence or a
                                list of sequences to classify.
        ---------------------   -------------------------------------------
        candidate_labels        Required string or list. The set of possible
                                class labels to classify each sequence into.
                                Can be a single label, a string of
                                comma-separated labels, or a list of labels.
        ---------------------   -------------------------------------------
        show_progress           optional Bool. If set to True, will display a
                                progress bar depicting the items processed so far.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        multi_class             Optional boolean. Whether or not multiple
                                candidate labels can be true.
                                Default value is set to False.
        ---------------------   -------------------------------------------
        hypothesis              Optional string. The template used to turn each
                                label into an NLI-style hypothesis. This template
                                must include a {} or similar syntax for the
                                candidate label to be inserted into the template.
                                Default value is set to `"This example is {}."`.
        =====================   ===========================================

        :return: a list of :obj:`dict`: Each result comes as a dictionary with the following keys:

            - **sequence** (:obj:`str`) -- The sequence for which this is the output.
            - **labels** (:obj:`List[str]`) -- The labels sorted by order of likelihood.
            - **scores** (:obj:`List[float]`) -- The probabilities for each of the labels.
        """
        results = []
        multi_class = kwargs.get("multi_class", False)
        hypothesis = kwargs.get("hypothesis", "This example is {}.")
        if not isinstance(text_or_list, (list, tuple)):
            text_or_list = [text_or_list]

        for i in progress_bar(range(len(text_or_list)), display=show_progress):
            result = self.model(
                text_or_list[i],
                candidate_labels,
                multi_class=multi_class,
                hypothesis_template=hypothesis,
            )
            results.append(result)
        return results



""" Defines store functions for working with Integrations.
"""

from ... import workforce
from .utils import add_features, update_features, remove_features, validate


def get_integration(project, integration_id=None):
    """Gets the identified Integration.  Exactly one integration_id should be provided.
    :param project:
    :param integration_id
    :returns: Integration
    """
    if integration_id:
        where = "{} = '{}'".format(
            project._integration_schema.integration_id, integration_id
        )
    else:
        where = "1=0"
    integrations = query_integrations(project, where)
    return integrations if integrations else None


def query_integrations(project, where):
    """Executes a query against the integrations table.
    :param project: The project in which to query integrations.
    :param where: An ArcGIS where clause.
    :returns: list of Integrations
    """
    features = project.integrations_table.query(where, return_all_records=True).features
    return [workforce.Integration(project, feature) for feature in features]


def add_integration(
    project,
    feature=None,
    integration_id=None,
    prompt=None,
    url_template=None,
    assignment_type=None,
):
    """
    Adds a new integration to the project
    """
    project._update_cached_objects()
    integration = workforce.Integration(
        project, feature, integration_id, prompt, url_template, assignment_type
    )
    return add_integrations(project, [integration])[0]


def add_integrations(project, integrations):
    """Adds Integrations to a project.

    :param project:
    :param integrations: list of Integrations
    :raises ValidationError: Indicates that one or more integrations failed validation.
    :raises ServerError: Indicates that the server rejected the integrations.
    """
    project._update_cached_objects()
    if integrations:
        use_global_ids = True
        for integration in integrations:
            validate(integration._validate)
            if integration.global_id is None:
                use_global_ids = False
                break

        features = [integration.feature for integration in integrations]
        add_features(project.integrations_table, features, use_global_ids)

    return integrations


def update_integration(
    project,
    integration,
    integration_id=None,
    prompt=None,
    url_template=None,
    assignment_type=None,
):
    """
    Updates an integration and submits changes to the server
    """
    project._update_cached_objects()
    if integration_id:
        integration.integration_id = integration_id
    if prompt:
        integration.prompt = prompt
    if url_template:
        integration.url_template = url_template
    if assignment_type:
        # GUIDs are cast to upper for internal storage purposes
        integration.assignment_type = assignment_type.upper()
    return update_integrations(project, [integration])[0]


def update_integrations(project, integrations):
    """Updates Integrations to a project.
    :param project:
    :param integrations: list of Integrations
    :raises ValidationError: Indicates that one or more dispatchers failed validation.
    :raises ServerError: Indicates that the server rejected the dispatchers.
    """
    project._update_cached_objects()
    if integrations:
        for integration in integrations:
            validate(integration._validate)
        features = [integration.feature for integration in integrations]
        update_features(project.integrations_table, features)

    return integrations


def delete_integrations(project, integrations):
    """Removes Integrations from the project.
    :param project:
    :param integrations: list of Integrations
    :raises ValidationError: Indicates that one or more integrations failed validation.
    :raises ServerError: Indicates that the server rejected the removal.
    """
    project._update_cached_objects()
    if integrations:
        features = [integration.feature for integration in integrations]
        remove_features(project.integrations_table, features)



from __future__ import annotations
import os
import uuid
import json
import shutil
import logging
import tempfile
import warnings
import concurrent.futures
from typing import Any
from arcgis.gis import GIS, Group, GroupManager
from arcgis.auth.tools import LazyLoader
from functools import lru_cache

from arcgis.gis.clone._base import BaseCloneGroup
from arcgis.gis.clone._utils import _zipdir

_arcgis = LazyLoader("arcgis")

_logger = logging.getLogger()


###########################################################################
class CloningJob:
    """
    A Single Group Cloning Job

    This class should not be created by users.
    """

    _future: concurrent.futures.Future
    _task: str

    # ---------------------------------------------------------------------
    def __init__(self, future: concurrent.futures.Future, task: str = "Group") -> None:
        self._future = future
        self._task = task

    # ---------------------------------------------------------------------
    def __str__(self) -> str:
        return f"< {self._task} Cloning Job: {self.status} >"

    # ---------------------------------------------------------------------
    def __repr__(self) -> str:
        return self.__str__()

    # ---------------------------------------------------------------------
    @property
    def status(self) -> bool:
        """checks if the job completed"""
        return self._future.done()

    # ---------------------------------------------------------------------
    def cancel(self) -> bool:
        """checks if the job completed"""
        return self._future.cancel()

    # ---------------------------------------------------------------------
    def result(self) -> Group:
        """returns a group"""
        return self._future.result()

    # ---------------------------------------------------------------------
    @property
    def running(self) -> bool:
        """checks if the job was cancelled"""
        return self._future.running()

    # ---------------------------------------------------------------------
    @property
    def cancelled(self) -> bool:
        """checks if the job was cancelled"""
        return self._future.cancelled()


###########################################################################
class GroupCloner(BaseCloneGroup):
    """
    The `GroupCloner` allows users to copy Groups from one organization to another.

    Once the groups are copied, the `clone_items` or Group Migration Manager can be
    used to add the data.

    """

    _group_source: Group = None
    _include_items: bool = None
    _gis: GIS = None
    _out_folder: str | None = None
    _tracker: dict[str, Any] | None = None
    _tp: concurrent.futures.ThreadPoolExecutor = None

    def __init__(
        self,
        *,
        gis: GIS | None = None,
    ) -> None:
        """initializer"""
        super()
        if gis:
            self._gis: GIS = gis
        elif _arcgis.env.active_env:
            self._gis = _arcgis.env.active_env
        else:
            raise ValueError("The `GIS` object is not defined.")

        self._tp = concurrent.futures.ThreadPoolExecutor(max_workers=10)

    # ---------------------------------------------------------------------
    def __str__(self) -> str:
        return "< Group Cloner >"

    # ---------------------------------------------------------------------
    def __repr__(self) -> str:
        return "< Group Cloner >"

    # ---------------------------------------------------------------------
    @lru_cache(maxsize=100)
    def _setup_group_project_folder(
        self, group_id: str, out_folder: str | None = None
    ) -> str:
        """creates a group project folder"""
        if out_folder:
            os.makedirs(out_folder, exist_ok=True)
        elif out_folder is None:
            out_folder = os.path.join(tempfile.gettempdir(), group_id)
            os.makedirs(out_folder, exist_ok=True)
        self._out_folder = out_folder
        return out_folder

    #  --------------------------------------------------------------------
    def _create_cloner_file(
        self, save_folder: str, path: str, package_name: str | None = None
    ) -> str:
        """Creates the Zipped compressed .GROUP_CLONER file."""
        if package_name is None:
            package_name = f"{uuid.uuid4().hex[:5]}.GROUP_CLONER"
        elif package_name.endswith(".GROUP_CLONER") == False:
            package_name += ".GROUP_CLONER"

        save_fp: str = os.path.join(save_folder, package_name)
        _zipdir(src=path, dst=save_folder, zip_name=package_name)

        if os.path.isfile(save_fp):
            return save_fp
        else:
            raise Exception(f"Could not save the the file: {save_fp}")
        return save_fp

    # ---------------------------------------------------------------------
    def _check_if_exists(self, group: Group) -> bool:
        """Checks if the group exists on the destination org"""
        results = self._gis.groups.search(f"id: {group.id}")
        return len(results) >= 1

    def _create_group_definition(
        self, groups: list[Group], save_folder: str, file_name: str
    ) -> dict[str, Any]:
        """creates the offline definition of the group"""
        group_defs: list[str] = []
        working_folder: str = os.path.join(
            tempfile.gettempdir(), uuid.uuid4().hex[:5]
        )  #  this is where we store the information on disk
        os.makedirs(
            save_folder, exist_ok=True
        )  # this ensures the save location of the cloner file exists
        group_json_file: str = os.path.join(working_folder, "group_def.json")
        fp: str = os.path.join(save_folder, file_name)

        lu: dict[str, Any] = {
            '{"itemTypes": "Application"}': "apps",
            '{"itemTypes": ""}': "none",
            '{"itemTypes": "CSV"}': "files",
            '{"itemTypes": "Web Map"}': "maps",
            '{"itemTypes": "Layer"}': "layers",
            '{"itemTypes": "Web Scene"}': "scenes",
            '{"itemTypes": "Locator Package"}': "tools",
        }
        for group in groups:
            group_thumbnail_folder: str = os.path.join(working_folder, group.id)
            os.makedirs(group_thumbnail_folder, exist_ok=True)

            display_settings: str | None = None
            if json.dumps(group.displaySettings) in lu:
                display_settings = lu[json.dumps(group.displaySettings)]
            tags: list[str] = group.tags
            tags.append(f"cloned_source_id: {group.id}")
            thumbnail: str = self._get_thumbnail(
                group=group, save_folder=group_thumbnail_folder
            )
            if getattr(group, "membershipAccess", None) != "collaboration":
                group_settings: dict[str, Any] = {
                    "title": group.title,
                    "tags": tags,
                    "description": group.description,
                    "snippet": group.snippet,
                    "thumbnail": thumbnail,
                    "is_invitation_only": group.isInvitationOnly,
                    "sort_field": group.sortField,
                    "sort_order": group.sortOrder,
                    "is_view_only": group.isViewOnly,
                    "auto_join": group.autoJoin,
                    "display_settings": display_settings,
                    "leaving_disallowed": group.leavingDisallowed,
                    "membership_access": getattr(group, "membershipAccess", None),
                    "hidden_members": getattr(group, "hiddenMembers", None),
                    "autojoin": group.autoJoin,
                }
                group_defs.append(group_settings)
            else:
                warnings.warn(
                    f"Collaboration groups are not able to be copied. {group.title}"
                )
        with open(group_json_file, "w") as writer:
            json.dump(group_defs, writer)
        fp = self._create_cloner_file(
            path=working_folder,
            save_folder=save_folder,
            package_name=file_name,
        )
        return fp

    # ---------------------------------------------------------------------
    def _process_offline(
        self,
        groups: list[Group],
        save_folder: str | None = None,
        file_name: str | None = None,
        **kwargs,
    ) -> str:
        if save_folder is None:
            save_folder = tempfile.gettempdir()
        if file_name is None:
            file_name = "GROUPSETTINGS"
        fp: str = self._create_group_definition(
            groups=groups, save_folder=save_folder, file_name=file_name
        )
        return fp

    # ---------------------------------------------------------------------
    def _process_online(self, params) -> Group:
        """
        Runs the clone process on a single group entry
        """
        group: Group = params["group"]
        working_folder: str = params["working_folder"]
        mgr: GroupManager = params["gis"].groups
        lu = {
            '{"itemTypes": "Application"}': "apps",
            '{"itemTypes": ""}': None,
            '{"itemTypes": "CSV"}': "files",
            '{"itemTypes": "Web Map"}': "maps",
            '{"itemTypes": "Layer"}': "layers",
            '{"itemTypes": "Web Scene"}': "scenes",
            '{"itemTypes": "Locator Package"}': "tools",
        }
        display_settings = None
        if json.dumps(group.displaySettings) in lu:
            display_settings = lu[json.dumps(group.displaySettings)]
        tags: list[str] = group.tags
        tags.append(f"cloned_source_id: {group.id}")
        thumbnail: str = self._get_thumbnail(group=group, save_folder=working_folder)
        group = mgr.create(
            title=group.title,
            tags=group.tags,
            description=group.description,
            snippet=group.snippet,
            thumbnail=thumbnail,
            is_invitation_only=group.isInvitationOnly,
            sort_field=group.sortField,
            sort_order=group.sortOrder,
            is_view_only=group.isViewOnly,
            auto_join=group.autoJoin,
            display_settings=display_settings,
            leaving_disallowed=group.leavingDisallowed,
            membership_access=getattr(group, "membershipAccess", None),
            hidden_members=getattr(group, "hiddenMembers", None),
            autojoin=group.autoJoin,
        )
        return group

    # ---------------------------------------------------------------------
    def _get_thumbnail(self, group: Group, save_folder: str) -> str:
        """
        downloads the thumbnail

        :returns: str
        """
        return group.download_thumbnail(save_folder=save_folder)

    # ---------------------------------------------------------------------
    def _run_ops(self, parameters: dict[str, Any], offline: bool = False) -> CloningJob:
        """runs the job asynchronously."""
        jobs = []

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as tp:
            if offline:
                future: concurrent.futures.Future = tp.submit(
                    self._process_offline, **parameters
                )
                jobs.append(CloningJob(future=future, task="Offline Group Cloning"))
            else:
                for key, params in parameters.items():
                    future: concurrent.futures.Future = tp.submit(
                        self._process_online, **{"params": params}
                    )

                    jobs.append(CloningJob(future=future))
                    del key, params
                tp.shutdown(wait=True)
        return jobs

    # ---------------------------------------------------------------------
    def load_offline_configuration(
        self, package: str
    ) -> list[concurrent.futures.Future]:
        """
        Loads the Group configuration file into the current active portal.
        """
        groups = []
        workspace: str = os.path.join(
            tempfile.gettempdir(), f"group_config_{uuid.uuid4().hex[:5]}"
        )
        os.makedirs(workspace, exist_ok=True)
        shutil.unpack_archive(filename=package, extract_dir=workspace, format="zip")
        settings_configuration = []
        with open(os.path.join(workspace, "group_def.json"), "r") as reader:
            settings_configuration = json.load(reader)
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as tp:
            for group in settings_configuration:
                if group["thumbnail"]:
                    thumbnail = os.path.join(
                        workspace,
                        os.path.basename(os.path.dirname(group["thumbnail"])),
                        os.path.basename(group["thumbnail"]),
                    )
                    group["thumbnail"] = thumbnail

                future = tp.submit(
                    self._load_offline_configuration,
                    **{
                        "payload": group,
                    },
                )
                groups.append(future)
            tp.shutdown(wait=True)
        return groups

    # ---------------------------------------------------------------------
    def _load_offline_configuration(self, payload: dict[str, Any]) -> Group | str:
        """
        Loads the Group configuration file into the current active portal.
        """
        gm = self._gis.groups
        try:
            return gm.create(**payload)
        except Exception as ex:
            return f"could not create the group {payload['title']} with error: {ex}"

    # ---------------------------------------------------------------------
    def clone(
        self,
        groups: list[Group],
        *,
        skip_existing: bool = True,
        offline: bool = False,
        save_folder: str | None = None,
        file_name: str = "group_list.json",
    ) -> list[CloningJob]:
        """
        Override the clone operation in order performs the cloning logic



        """
        self._tracker: dict[str, Any] = {}
        ## 1). Check Existance and setup project
        ##
        if offline:
            params = {
                "groups": groups,
                "skip_existing": skip_existing,
                "save_folder": save_folder,
                "file_name": file_name,
            }
            return self._run_ops(parameters=params, offline=True)
        else:
            for group in groups:
                if (
                    self._check_if_exists(group=group) == False and skip_existing
                ) or skip_existing == False:
                    self._tracker[group.id] = {
                        "group": group,
                        "gis": self._gis,
                        "exists": self._check_if_exists(group),
                        "working_folder": self._setup_group_project_folder(
                            group_id=group.id
                        ),
                    }
            return self._run_ops(parameters=self._tracker)



from typing import Tuple
from requests.auth import AuthBase
from ._schain import SupportMultiAuth
from ..tools._lazy import LazyLoader
from ..tools import parse_url, assemble_url

os = LazyLoader("os")
tempfile = LazyLoader("tempfile")
_dt = LazyLoader("datetime")
requests = LazyLoader("requests")


###########################################################################
class EsriPKIAuth(AuthBase, SupportMultiAuth):
    """Handles PKI authentication when tokens are needed"""

    _token_url = None
    _server_log = None
    _server_log_count = None
    _tokens = None
    _session = None

    def __init__(
        self,
        cert: Tuple[str],
        referer: str = None,
        verify_cert: bool = True,
        **kwargs,
    ):
        self._server_log = {}
        self._server_log_time = {}
        self._tokens = {}
        self._token_url = None
        self.verify_cert = verify_cert
        self.cert = cert
        self.auth = kwargs.pop("auth", None)
        if referer is None:
            self.referer = "http"
        else:
            self.referer = referer
        self._session = kwargs.pop("session", requests.Session())
        self._proxies = kwargs.pop("proxies", None)

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    # ----------------------------------------------------------------------
    @property
    def token(self) -> str:
        """
        returns the authentication token
        """
        return None

    # ----------------------------------------------------------------------
    def generate_portal_server_token(self, r, **kwargs):
        """generates a server token using Portal token"""
        parsed = parse_url(r.url)
        server_url = assemble_url(parsed)
        if (
            r.text.lower().find("invalid token") > -1
            or r.text.lower().find("token required") > -1
            or r.text.lower().find("token not found") > -1
        ) or server_url in self._server_log:
            expiration = 16000
            postdata = {
                "request": "getToken",
                "serverURL": server_url,
                "referer": self.referer or "http",
                "f": "json",
            }
            if expiration:
                postdata["expiration"] = expiration
            if server_url in self._server_log:
                token_url = self._server_log[server_url]
                self._server_log_time[server_url] = _dt.datetime.now() + _dt.timedelta(
                    minutes=expiration
                )
            else:
                info = self._session.get(
                    server_url + "/rest/info?f=json",
                    cert=(self.cert[0], self.cert[1]),
                    verify=self.verify_cert,
                    auth=self.auth,
                    proxies=self._proxies,
                ).json()
                token_url = info["authInfo"]["tokenServicesUrl"]
                self._server_log[server_url] = token_url
            if server_url in self._tokens:
                if _dt.datetime.now() >= self._server_log_time[server_url]:
                    del self._tokens[server_url]
                    return self.generate_portal_server_token(r)
                token_str = self._tokens[server_url]
            else:
                token = self._session.post(
                    token_url,
                    data=postdata,
                    cert=(self.cert[0], self.cert[1]),
                    auth=self.auth,
                    verify=self.verify_cert,
                    proxies=self._proxies,
                )
                token_str = token.json().get("token", None)
                if token_str is None:
                    return r
                self._tokens[server_url] = token_str
            # Recreate the request with the token
            #
            r.content
            r.raw.release_conn()
            r.request.headers["referer"] = self.referer or "http"
            r.request.headers["X-Esri-Authorization"] = f"Bearer {token_str}"
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self.referer or "http"
            _r.headers["X-Esri-Authorization"] = f"Bearer {token_str}"
            _r.history.append(r)
            return _r
        return r

    # ----------------------------------------------------------------------
    def __call__(self, r):
        if self.auth:
            self.auth.__call__(r)
        r.register_hook("response", self.generate_portal_server_token)
        return r



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model settings
model = dict(
    type="RetinaNet",
    pretrained="torchvision://resnet50",
    backbone=dict(
        type="ResNet",
        depth=50,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        norm_eval=True,
        style="pytorch",
    ),
    neck=dict(
        type="FPN",
        in_channels=[256, 512, 1024, 2048],
        out_channels=256,
        start_level=1,
        add_extra_convs="on_input",
        num_outs=5,
    ),
    bbox_head=dict(
        type="RetinaHead",
        num_classes=80,
        in_channels=256,
        stacked_convs=4,
        feat_channels=256,
        anchor_generator=dict(
            type="AnchorGenerator",
            octave_base_scale=4,
            scales_per_octave=3,
            ratios=[0.5, 1.0, 2.0],
            strides=[8, 16, 32, 64, 128],
        ),
        bbox_coder=dict(
            type="DeltaXYWHBBoxCoder",
            target_means=[0.0, 0.0, 0.0, 0.0],
            target_stds=[1.0, 1.0, 1.0, 1.0],
        ),
        loss_cls=dict(
            type="FocalLoss", use_sigmoid=True, gamma=2.0, alpha=0.25, loss_weight=1.0
        ),
        loss_bbox=dict(type="L1Loss", loss_weight=1.0),
    ),
    # training and testing settings
    train_cfg=dict(
        assigner=dict(
            type="MaxIoUAssigner",
            pos_iou_thr=0.5,
            neg_iou_thr=0.4,
            min_pos_iou=0,
            ignore_iof_thr=-1,
        ),
        allowed_border=-1,
        pos_weight=-1,
        debug=False,
    ),
    test_cfg=dict(
        nms_pre=1000,
        min_bbox_size=0,
        score_thr=0.05,
        nms=dict(type="nms", iou_threshold=0.5),
        max_per_img=100,
    ),
)



from __future__ import annotations

try:
    import ujson as json
except ImportError:
    import json
import time
import concurrent.futures
from functools import lru_cache
from arcgis._impl.common._mixins import PropertyMap
from typing import Dict, Any, Optional, List
from arcgis.gis import GIS
import datetime as _dt


###########################################################################
class GPJobManager:
    """
    The jobs resource provides access to operations that locate and monitor
    current asynchronous jobs being run by a geoprocessing service. From
    the jobs resource, you can query for jobs using filters such as the
    start and end time for the job, the job's status, or the username of
    the user who submitted the job.
    """

    _gis = None
    _url = None
    _properties = None

    # ----------------------------------------------------------------------
    def __init__(self, url: str, gis: GIS):
        self._url = url
        self._gis = gis

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> dict:
        params = {"f": "json"}
        return self._gis._con.get(self._url, params)

    # ----------------------------------------------------------------------
    def get_job(self, job_id: str) -> dict:
        """
        An individual job resource returns information about an asynchronous
        job, either currently running or completed, for a geoprocessing service.
        """
        url = self._url + f"/{job_id}"
        params = {"f": "json"}
        return self._gis._con(url, params)

    # ----------------------------------------------------------------------
    def query(
        self,
        status: list[str],
        start_time: _dt.datetime | None = None,
        end_time: _dt.datetime | None = None,
        username: str | None = None,
        number: int = 10,
    ) -> dict:
        """
        The query operation allows you to query jobs pertaining to a geoprocessing service.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        status              Required List[String]. The current status of a job. The value set with
                            this parameter will be used to filter the jobs by that set job
                            status.

                            Values: `esriJobNew`, `esriJobSubmitted`, `esriJobExecuting`, `esriJobSucceeded`, `esriJobFailed`, `esriJobCancelling`, `esriJobCancelled`, `esriJobWaiting`
        ---------------     --------------------------------------------------------------------
        start_time          Optional datetime.datetime. The earliest time to query.
        ---------------     --------------------------------------------------------------------
        end_time            Optional datetime.datetime. The most recent time to query. If unspecified, the current time will be used. If you specify a value for this parameter, you must also specify a startTime value.
        ---------------     --------------------------------------------------------------------
        username            Optional String. The name of the user who submitted the job.
        ---------------     --------------------------------------------------------------------
        number              Optional Integer. The number of jobs to display in the response. The default value is 10.
        ===============     ====================================================================

        :returns: dict

        """
        status_message = {
            "esriJobWaiting",
            "esriJobFailed",
            "esriJobSubmitted",
            "esriJobCancelled",
            "esriJobNew",
            "esriJobExecuting",
            "esriJobSucceeded",
            "esriJobCancelling",
        }
        url = self._url + "/query"
        params = [
            ("f", "json"),
            ("start", 1),
            ("number", number),
        ]
        if start_time and isinstance(start_time, _dt.datetime):
            params.append(("startTime", int(start_time.timestamp())))
        if end_time and isinstance(end_time, _dt.datetime) and start_time:
            params.append(("endTime", int(end_time.timestamp())))

        for s in status:
            params.append(("jobStatus", s))
        if username:
            params.append(("userName", username))
        results = {}
        res = self._gis._con.get(url, params)
        results.update(res.get("results", {}))
        while res["nextStart"] > -1 or len(res["results"]) > 0:
            if res["nextStart"] == -1:
                break
            params = [
                ("f", "json"),
                ("start", res["nextStart"]),
                ("number", number),
            ]
            if start_time and isinstance(start_time, _dt.datetime):
                params.append(("startTime", int(start_time.timestamp())))
            if end_time and isinstance(end_time, _dt.datetime) and start_time:
                params.append(("endTime", int(end_time.timestamp())))

            for s in status:
                params.append(("jobStatus", s))
            if username:
                params.append(("userName", username))
            res = self._gis._con.get(url, params)
            results.update(res.get("results", {}))
        return results


###########################################################################
class KubeService(object):
    """
    A service exposes GIS resources like maps, rasters, locators,
    geodatabases, and so forth through REST and SOAP interfaces. The type
    of the service is often dictated by the type of resources being
    published. In addition to accessing the underlying resource, a GIS
    service can expose additional capabilities called extensions (or server
    object extensions). Extensions are packages of custom functionality that
    can perform business logic or expose the service through additional
    formats or protocols.
    """

    _url = None
    _con = None
    _gis = None
    _properties = None

    def __init__(self, url, gis):
        self._url = url
        self._gis = gis
        self._con = gis._con

    def __str__(self) -> str:
        return f"< Kubernetes Admin Service: {self.properties['serviceName']} >"

    def __repr__(self) -> str:
        return self.__str__()

    # ----------------------------------------------------------------------
    def _status(self, url: str, gis: GIS) -> dict:
        job = self._gis._con.get(url, {"f": "json"})
        i = 0
        while job["status"] == "EXECUTING":
            job = self._gis._con.get(url, {"f": "json"})
            time.sleep(i)
            if i > 5:
                i = 4
            i += 1
            if job["status"].lower() == "completed":
                break
        return job

    # ----------------------------------------------------------------------
    def change_provider(self, provider: str) -> bool:
        """
        This operation is used to update an individual service to use either
        a dedicated or shared instance type. When a qualified service is
        published, the service is automatically set to use shared instances.

        When using this operation, services may populate other provider types
        as values for the provider parameter, such as ArcObjects and SDS.
        While these are valid provider types, this operation does not support
        changing the provider of such services to either ArcObjects11 or
        DMaps. Services with either ArcObjects or SDS as their provider will
        not be able to change their instance type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        provider            Required String. Specifies the service instance as either a shared
                            (DMaps) or dedicated (ArcObjects11) instance type. These values are case-sensitive.

                            values: DMaps, ArcObjects11
        ===============     ====================================================================


        :return: boolean
        """
        lu = {"dmaps": "DMaps", "arcobjects11": "ArcObjects11"}
        provider = lu[provider.lower()]
        params = {"f": "json", "provider": provider}
        url = "{self._url}/changeProvider"
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    def delete(self) -> bool:
        """
        Removes the service from the hosting server

        :return: bool

        """
        url = f"{self._url}/delete"
        params = {"f": "json"}
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    @property
    def status(self) -> dict:
        """
        This resource provides the configured and current status of a GIS
        service. The configured status represents the state of the resource
        as you have configured it to be. For example, starting a service
        would set its configured status to be STARTED. However, it is
        possible that the configured state may not match the actual state
        of the resource. The realTimeState property represents the actual
        state of a service.
        """
        url = f"{self._url}/status"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ----------------------------------------------------------------------
    def start(self) -> bool:
        """
        Starts the service

        :return: bool
        """

        url = f"{self._url}/start"
        params = {"f": "json"}
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def stop(self) -> bool:
        """
        Stops the service

        :return: bool
        """

        url = f"{self._url}/stop"
        params = {"f": "json"}
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def restart(self) -> bool:
        """
        Recycles the current service
        :return: Bool
        """
        self.stop()
        return self.start()

    # ----------------------------------------------------------------------
    @property
    def scaling(self) -> Dict[str, Any]:
        """
        This resource returns the scaling and resource allocation for a
        specific GIS service microservice. When used to update the service,
        it updates the scaling (replicas min and max) and resource allocation
        (cpuMin, cpuMax, memoryMin, memoryMax).

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        value               Required Dict[str, Any]. The service scaling properties.
        ===============     ====================================================================

        :return: Dict[str, Any]

        """
        url = f"{self._url}/scaling"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ----------------------------------------------------------------------
    @scaling.setter
    def scaling(self, value: Dict[str, Any]):
        """
        See main ``scaling`` property docstring
        """
        future = True
        url = f"{self._url}/scaling/edit"
        params = {"f": "json", "serviceScalingSpec": value, "async": True}
        res = self._con.post(url, params)
        job_url = res.get("jobsUrl", None)
        values = {"url": job_url, "gis": self._gis}
        tp = concurrent.futures.ThreadPoolExecutor(1)
        try:
            future = tp.submit(fn=self._status, **values)
        except:
            future = tp.submit(self._status, **values)
        tp.shutdown(False)
        future.result()

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> PropertyMap:
        """
        To edit a service, you need to submit the complete JSON representation of
        the service, which includes the updates to the service properties. Editing
        a service can cause the service to be restarted with updated properties.

        The edit settings of a service contains the following four sections:

        + Service Description Properties - Common properties that are shared by all services. These properties typically identify a specific service.
        + Service Framework Properties - Properties targets towards the framework that hosts the GIS service. They define the life cycle and load balancing of the service.
        + Service Type Properties - Properties targeted towards the core service type as seen by the server administrator. Since these properties are associated with a server object, they vary across the service types.
        + Extension Properties - Represent the extensions that are enabled on the service.

        :return: dict
        """
        if self._properties is None:
            url = self._url
            params = {"f": "json"}
            res = self._con.get(url, params)
            self._properties = PropertyMap(res)
        return self._properties

    # ----------------------------------------------------------------------
    @properties.setter
    def properties(self, properties: dict):
        """
        See main ``properties`` property docstring.
        """
        historic = dict(self.properties)
        historic.update(properties)
        params = {"f": "json", "service": historic}
        url = f"{self._url}/edit"
        res = self._con.post(url, params)
        self._properties = None
        if "status" in res and not res["status"] == "success":
            raise Exception(str(res))

    # ----------------------------------------------------------------------
    @property
    def jobs(self) -> GPJobManager | None:
        """
        The jobs resource provides access to operations that locate and
        monitor current asynchronous jobs being run by a geoprocessing
        service. From the jobs resource, you can query for jobs using
        filters such as the start and end time for the job, the job's
        status, or the username of the user who submitted the job.

        :returns: JobManager
        """
        if self._url.lower().find(".gpserver") > -1:
            return GPJobManager(url=f"{self._url}/jobs", gis=self._gis)
        return None


###########################################################################
class ServicesManager(object):
    """
    The `ServicesManager` acts as the root folder and container for all
    sub-folders and GIS services for your deployment. You can create a
    new sub-folder by using the Create Folder operation as well as a new
    GIS service by using the Create Service method.
    """

    _gis = None
    _url = None
    _folder = None
    _types = None
    _properties = None

    def __init__(self, url: str, gis: "GIS"):
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    @property
    def services_properties(self) -> dict:
        """
        This resource is used to provide default settings for new services
        when they are published to the server. You can use the update
        operation to change the default settings. However, updating the
        default properties for services won't change the properties of any
        pre-existing services in your organization. To update these services,
        you must edit the individual service's properties.

        :returns: dict

        """
        return self._con.get(self._url + "/properties", {"f": "json"})

    # ----------------------------------------------------------------------
    @services_properties.setter
    def services_properties(self, properties: dict):
        """
        See main ``service_properties`` docstring
        """
        res = self._con.post(
            self._url + "/properties", {"f": "json", "properties": properties}
        )
        if res["status"] != "success":
            raise Exception(res)

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> dict:
        """
        The proeprties of the manager.

        :return: dict
        """
        url = self._url
        params = {"f": "json"}
        self._properties = self._con.get(url, params)
        return self._properties

    # ----------------------------------------------------------------------
    @property
    def folder(self):
        """returns the current folder"""
        return self._folder or "/"

    # ----------------------------------------------------------------------
    @folder.setter
    def folder(self, folder: str):
        """get/set the current folder"""
        if folder is None:
            self._folder = "/"
        elif folder in self.folders:
            self._folder = folder
        else:
            raise Exception("Folder does not exist.")

    # ----------------------------------------------------------------------
    def list(self, folder: str = None) -> list:
        """lists the services in the current folder"""
        services = []

        if folder:
            url = f"{self._url}/{folder}"
        else:
            url = self._url
        res = self._con.get(url, {"f": "json"})
        if "services" in res:
            return [
                KubeService(
                    url=f"{url}/{service['serviceName']}.{service['type']}",
                    gis=self._gis,
                )
                for service in res["services"]
            ]
        return []

    # ----------------------------------------------------------------------
    @property
    def folders(self) -> list:
        """
        returns a list of folder names

        :return: List[str]

        """
        return self.properties.get("folders", [])

    # ----------------------------------------------------------------------
    @property
    def types(self) -> dict:
        """
        This resource provides metadata about all service types and extensions that can be enabled.

        :return: Dict

        """
        if self._types is None:
            self._types = self._con.get(f"{self._url}/types", {"f": "json"})
        return self._types

    # ----------------------------------------------------------------------
    @property
    def extensions(self) -> dict:
        """
        Returns a collection of all the custom server object extensions that have been uploaded
        and registered with the deployment. A .soe file is a container of one or more server object
        extensions.

        :return: Dict
        """
        return self._con.get(f"{self._url}/types/extensions", {"f": "json"})

    # ----------------------------------------------------------------------
    @property
    def providers(self) -> dict:
        """
        returns the supported provider types for the GIS services in your organization.
        """
        return self._con.get(f"{self._url}/types/providers", {"f": "json"})

    # ----------------------------------------------------------------------
    def exists(
        self, *, service_name: str = None, folder: str = None, service_type: str = None
    ) -> dict:
        """
        This operation checks if a folder or service exists on the server.

        :return: dict
        """
        params = {
            "f": "json",
            "folderName": folder,
            "serviceName": service_name,
            "type": service_type,
        }
        for key in list(params.keys()):
            if params[key] is None:
                del params[key]

        url = f"{self._url}/exists"
        res = self._con.post(url, params)
        return res

    # ----------------------------------------------------------------------
    def can_create(
        self,
        service_type: str,
        *,
        folder: str = None,
        service: dict = None,
        options: dict = None,
    ) -> bool:
        """
        Checks if a service can be generated. It is recommended that the user
        check if the service can be created before calling `create_service`.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        service_type        Required String.  The type of service to create.
        ---------------     --------------------------------------------------------------------
        folder              Optional String. The location to create the service in.  If the `folder`
                            if set on the `ServicesManager`, the `folder` parameter will override
                            the save location. The folder must exist before calling this method.
        ---------------     --------------------------------------------------------------------
        service             Optional Dict. The service configuration in JSON format.
        ---------------     --------------------------------------------------------------------
        options             Optional Dict. Provides additional information about the service, such as whether it is a hosted service.
        ===============     ====================================================================

        """
        url = f"{self._url}/canCreateService"
        params = {
            "f": "json",
            "folderName": folder,
            "serviceType": service_type,
            "service": service,
            "options": options,
        }
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    def create_service(
        self,
        service_json: dict = None,
        *,
        input_upload_id: str = None,
        folder: str = None,
        scaling_spec: dict = None,
    ) -> bool:
        """
        Creates a new GIS service in a folder (either the root or a sub-folder) by
        submitting a JSON representation of the service to this operation.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        service_json        Required dict. The JSON representation of the service being created.
        ---------------     --------------------------------------------------------------------
        folder              Optional String. The location to create the service in.  If the `folder`
                            if set on the `ServicesManager`, the `folder` parameter will override
                            the save location. The folder must exist before calling this method.
        ---------------     --------------------------------------------------------------------
        input_upload_id     Optional String. The upload ID for a service definition that
                            contains information about service properties, capabilities, and the
                            service type.
        ---------------     --------------------------------------------------------------------
        scaling_spec        Optional dict. The service scaling properties, represented as a JSON
                            object. See: https://developers.arcgis.com/rest/enterprise-administration/enterprise/create-service.htm for more information.
        ===============     ====================================================================

        :return: Bool

        """
        if isinstance(service_json, dict):
            service_json = json.dumps(service_json)

        if folder:
            url = f"{self._url}/{folder}/createService"
        elif folder is None and (self.folder or self.folder != "/"):
            url = f"{self._url}/{self.folder}/createService"
        else:
            url = f"{self._url}/createService"
        params = {"f": "json", "serviceJson": service_json}
        if input_upload_id:
            params["inputUploadId"] = input_upload_id

        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    def _delete_services(self, services: list[dict[str, Any]]) -> bool:
        """
        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        services            Required list[dict[str, Any]]. A list of services, represented as a
                            list of dictionaries containing the folders containing the services,
                            the services' names, and the service types.
                            Example
                            ```{
                               "folderName": "",
                               "serviceName": "SampleWorldCities",
                               "type": "MapServer"
                              },
                              {
                               "folderName": "Planning",
                               "serviceName": "FireHydrants",
                               "type": "FeatureServer"
                            }```

        ---------------     --------------------------------------------------------------------
        future              Optional Boolean.Indicates whether to process the operation in
                            synchronous or asynchronous mode. The default value is False.
                            When set to true, the operation returns a Future object that can be
                            used as a request to retrieve the job status for the operation.
        ===============     ====================================================================

        :returns: bool | concurrent.futures.Future
        """
        url = self._url + "/deleteServices"
        services = ",".join([json.dumps(service) for service in services])
        params = {"f": "json", "services": services, "async": False}
        res = self._con.post(url, params)
        return res.get("success", False)

    # ----------------------------------------------------------------------
    def create_folder(self, folder: str) -> bool:
        """
        Creates a folder on the hosting server

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        folder              Required string. Name of the folder.
        ===============     ====================================================================

        :return: boolean
        """
        url = self._url + "/createFolder"
        params = {"f": "json", "folderName": folder}
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    def _delete_folder(self, folder: str) -> bool:
        """
        Removes a folder on the hosting server

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        folder              Required string. Name of the folder.
        ===============     ====================================================================

        :return: boolean
        """
        params = {"f": "json"}
        if folder in self.folders:
            u_url = self._url + "/%s/deleteFolder" % folder
            res = self._con.post(path=u_url, postdata=params, try_json=False)
            return not folder in self.folders
            # if 'status' in res:
            #    return res['status'] == 'success'
            # return res
        else:
            return False

    # ----------------------------------------------------------------------
    def _status(self, url: str, gis: GIS) -> dict:
        job = self._gis._con.get(url, {"f": "json"})
        i = 0
        while job["status"] == "EXECUTING":
            job = self._gis._con.get(url, {"f": "json"})
            time.sleep(i)
            if i > 5:
                i = 4
            i += 1
            if job["status"].lower() == "completed":
                break
        return job

    # ----------------------------------------------------------------------
    def refresh_auto_deployment(self, future: bool = False) -> bool:
        """
        This operation auto-deploys the System or Utility services if they failed to be deployed
        during site creation. This operation should only be performed if either the System or
        Utility service fails to be created with the site.

        :return: Boolean

        """
        url = f"{self._url}/refreshAutodeployedServices"
        params = {"f": "json", "async": True}
        res = self._con.post(url, params)
        job_url = res.get("jobsUrl", None)
        values = {"url": job_url, "gis": self._gis}
        tp = concurrent.futures.ThreadPoolExecutor(1)
        try:
            future = tp.submit(fn=self._status, **values)
        except:
            future = tp.submit(self._status, **values)
        tp.shutdown(False)
        if future:
            return future
        return future.result()



# Note that these may change and need to be updated periodically
# https://developers.arcgis.com/javascript/latest/api-reference/esri-Map.html#basemap -> NO API KEY MAPS
# https://devtopia.esri.com/WebGIS/arcgis-js-api/blob/4master/esri/support/basemapDefinitions.ts -> for the json structure

basemap_dict = {
    "satellite": [
        {
            "id": "satellite-base-layer",
            "url": "https://services.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer",
            "layerType": "ArcGISTiledMapServiceLayer",
            "title": "World Imagery",
            "showLegend": False,
            "visibility": True,
            "opacity": 1,
        }
    ],
    "hybrid": [
        {
            "id": "hybrid-base-layer",
            "url": "https://services.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer",
            "layerType": "ArcGISTiledMapServiceLayer",
            "title": "World Imagery",
            "showLegend": False,
            "visibility": True,
            "opacity": 1,
        },
        {
            "id": "hybrid-reference-layer",
            "styleUrl": "https://www.arcgis.com/sharing/rest/content/items/30d6b8271e1849cd9c3042060001f425/resources/styles/root.json",
            "layerType": "VectorTileLayer",
            "title": "Hybrid Reference Layer",
            "isReference": True,
            "showLegend": False,
            "visibility": True,
            "opacity": 1,
        },
    ],
    "terrain": [
        {
            "id": "terrain-base-layer",
            "url": "https://services.arcgisonline.com/ArcGIS/rest/services/World_Terrain_Base/MapServer",
            "layerType": "ArcGISTiledMapServiceLayer",
            "title": "World Terrain Base",
            "showLegend": False,
            "visibility": True,
            "opacity": 1,
        },
        {
            "id": "terrain-reference-layer",
            "url": "https://services.arcgisonline.com/ArcGIS/rest/services/Reference/World_Reference_Overlay/MapServer",
            "layerType": "ArcGISTiledMapServiceLayer",
            "title": "World Reference Overlay",
            "isReference": True,
            "showLegend": False,
            "visibility": True,
            "opacity": 1,
        },
    ],
    "oceans": [
        {
            "id": "oceans-base-layer",
            "url": "https://services.arcgisonline.com/arcgis/rest/services/Ocean/World_Ocean_Base/MapServer",
            "layerType": "ArcGISTiledMapServiceLayer",
            "title": "World Ocean Base",
            "showLegend": False,
            "visibility": True,
            "opacity": 1,
        },
        {
            "id": "oceans-reference-layer",
            "url": "https://services.arcgisonline.com/arcgis/rest/services/Ocean/World_Ocean_Reference/MapServer",
            "layerType": "ArcGISTiledMapServiceLayer",
            "title": "World Ocean Reference",
            "isReference": True,
            "showLegend": False,
            "visibility": True,
            "opacity": 1,
        },
    ],
    "osm": [
        {
            "id": "osm-base-layer",
            "layerType": "OpenStreetMap",
            "title": "Open Street Map",
            "showLegend": False,
            "visibility": True,
            "opacity": 1,
        }
    ],
    "dark-gray-vector": [
        {
            "id": "dark-gray-base-layer",
            "styleUrl": "https://www.arcgis.com/sharing/rest/content/items/5e9b3685f4c24d8781073dd928ebda50/resources/styles/root.json",
            "layerType": "VectorTileLayer",
            "title": "Dark Gray Base",
            "visibility": True,
            "opacity": 1,
        },
        {
            "id": "dark-gray-reference-layer",
            "styleUrl": "https://www.arcgis.com/sharing/rest/content/items/747cb7a5329c478cbe6981076cc879c5/resources/styles/root.json",
            "layerType": "VectorTileLayer",
            "title": "Dark Gray Reference",
            "isReference": True,
            "visibility": True,
            "opacity": 1,
        },
    ],
    "gray-vector": [
        {
            "id": "gray-base-layer",
            "styleUrl": "https://www.arcgis.com/sharing/rest/content/items/291da5eab3a0412593b66d384379f89f/resources/styles/root.json",
            "layerType": "VectorTileLayer",
            "title": "Light Gray Base",
            "visibility": True,
            "opacity": 1,
        },
        {
            "id": "gray-reference-layer",
            "styleUrl": "https://www.arcgis.com/sharing/rest/content/items/1768e8369a214dfab4e2167d5c5f2454/resources/styles/root.json",
            "layerType": "VectorTileLayer",
            "title": "Light Gray Reference",
            "isReference": True,
            "visibility": True,
            "opacity": 1,
        },
    ],
    "streets-vector": [
        {
            "id": "streets-vector-base-layer",
            "styleUrl": "##cdn.arcgis.com/sharing/rest/content/items/de26a3cf4cc9451298ea173c4b324736/resources/styles/root.json",
            "layerType": "VectorTileLayer",
            "title": "World Streets",
            "visibility": True,
            "opacity": 1,
        }
    ],
    "topo-vector": [
        {
            "id": "world-hillshade-layer",
            "url": "https://services.arcgisonline.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer",
            "layerType": "ArcGISTiledMapServiceLayer",
            "title": "World Hillshade",
            "showLegend": False,
            "visibility": True,
            "opacity": 1,
        },
        {
            "id": "topo-vector-base-layer",
            "styleUrl": "https://www.arcgis.com/sharing/rest/content/items/7dc6cea0b1764a1f9af2e679f642f0f5/resources/styles/root.json",
            "layerType": "VectorTileLayer",
            "title": "World Topo",
            "visibility": True,
            "opacity": 1,
        },
    ],
    "streets-night-vector": [
        {
            "id": "streets-night-vector-base-layer",
            "styleUrl": "https://www.arcgis.com/sharing/rest/content/items/86f556a2d1fd468181855a35e344567f/resources/styles/root.json",
            "layerType": "VectorTileLayer",
            "title": "World Streets Night",
            "visibility": True,
            "opacity": 1,
        }
    ],
    "streets-relief-vector": [
        {
            "id": "world-hillshade-layer",
            "url": "//services.arcgisonline.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer",
            "layerType": "ArcGISTiledMapServiceLayer",
            "title": "World Hillshade",
            "showLegend": False,
            "visibility": True,
            "opacity": 1,
        },
        {
            "id": "streets-relief-vector-base-layer",
            "styleUrl": "//www.arcgis.com/sharing/rest/content/items/b266e6d17fc345b498345613930fbd76/resources/styles/root.json",
            "title": "World Streets Relief",
            "layerType": "VectorTileLayer",
            "visibility": True,
            "opacity": 1,
        },
    ],
    "streets-navigation-vector": [
        {
            "id": "streets-navigation-vector-base-layer",
            "styleUrl": "https://www.arcgis.com/sharing/rest/content/items/63c47b7177f946b49902c24129b87252/resources/styles/root.json",
            "layerType": "VectorTileLayer",
            "title": "World Streets Navigation",
            "visibility": True,
            "opacity": 1,
        }
    ],
    "arcgis-imagery": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Imagery",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/World_Imagery/MapServer",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Imagery:Labels",
            "title": "Hybrid Reference Layer",
            "isReference": True,
        },
    ],
    "arcgis-imagery-standard": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Imagery",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/World_Imagery/MapServer",
        }
    ],
    "arcgis-imagery-labels": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Imagery:Labels",
            "title": "Hybrid Reference Layer",
            "isReference": True,
        }
    ],
    "arcgis-light-gray": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:LightGray:Base",
            "title": "Light Gray Canvas Base",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:LightGray:Labels",
            "title": "Light Gray Canvas Labels",
            "isReference": True,
        },
    ],
    "arcgis-dark-gray": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:DarkGray:Base",
            "title": "Dark Gray Canvas Base",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:DarkGray:Labels",
            "title": "Dark Gray Canvas Labels",
            "isReference": True,
        },
    ],
    "arcgis-navigation": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Navigation",
            "title": "World Navigation Map",
        }
    ],
    "arcgis-navigation-night": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:NavigationNight",
            "title": "World Navigation Map (Dark Mode)",
        }
    ],
    "arcgis-streets": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Streets",
            "title": "World Street Map",
        }
    ],
    "arcgis-streets-night": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:StreetsNight",
            "title": "World Street Map (Night)",
        }
    ],
    "arcgis-streets-relief": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Hillshade",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:StreetsRelief:Base",
            "title": "World Street Map (with Relief)",
        },
    ],
    "arcgis-topographic": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Hillshade",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Topographic:Base",
            "title": "World Topographic Map",
        },
    ],
    "arcgis-oceans": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Ocean Base",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/Ocean/World_Ocean_Base/MapServer",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Oceans:Labels",
            "title": "World Ocean Reference",
            "isReference": True,
        },
    ],
    "osm-standard": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/OSM:Standard",
            "title": "OpenStreetMap",
        }
    ],
    "osm-standard-relief": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Hillshade",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer",
        },
        {
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/OSM:StandardRelief:Base",
            "layerType": "VectorTileLayer",
            "title": "OpenStreetMap Relief Base",
        },
    ],
    "osm-streets": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/OSM:Streets",
            "title": "OpenStreetMap (Streets)",
        }
    ],
    "osm-streets-relief": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Hillshade",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer",
        },
        {
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/OSM:StreetsRelief:Base",
            "layerType": "VectorTileLayer",
            "title": "OpenStreetMap Relief Base",
        },
    ],
    "osm-light-gray": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/OSM:LightGray:Base",
            "title": "OSM (Light Gray Base)",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/OSM:LightGray:Labels",
            "title": "OSM (Light Gray Reference)",
            "isReference": True,
        },
    ],
    "osm-dark-gray": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/OSM:DarkGray:Base",
            "title": "OSM (Dark Gray Base)",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/OSM:DarkGray:Labels",
            "title": "OSM (Dark Gray Reference)",
            "isReference": True,
        },
    ],
    "arcgis-terrain": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Hillshade",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Terrain:Base",
            "title": "World Terrain Base",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Terrain:Detail",
            "title": "World Terrain Reference",
            "isReference": True,
        },
    ],
    "arcgis-community": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Community",
            "title": "Community",
        }
    ],
    "arcgis-charted-territory": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Hillshade",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:ChartedTerritory:Base",
            "title": "Charted Territory",
        },
    ],
    "arcgis-colored-pencil": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:ColoredPencil",
            "title": "Colored Pencil",
        }
    ],
    "arcgis-nova": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Nova",
            "title": "Nova",
        }
    ],
    "arcgis-modern-antique": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Hillshade",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer",
        },
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:ModernAntique:Base",
            "title": "Modern Antique",
        },
    ],
    "arcgis-midcentury": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Midcentury",
            "title": "Mid-Century",
        }
    ],
    "arcgis-newspaper": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemaps-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:Newspaper",
            "title": "Newspaper",
        }
    ],
    "arcgis-hillshade-light": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Hillshade",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer",
        }
    ],
    "arcgis-hillshade-dark": [
        {
            "layerType": "ArcGISTiledMapServiceLayer",
            "showLegend": False,
            "title": "World Hillshade (Dark)",
            "url": "https://ibasemaps-api.arcgis.com/arcgis/rest/services/Elevation/World_Hillshade_Dark/MapServer",
        }
    ],
    "arcgis-human-geography": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemapsdev-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:HumanGeography",
            "title": "HumanGeography",
        }
    ],
    "arcgis-human-geography-dark": [
        {
            "layerType": "VectorTileLayer",
            "styleUrl": "https://basemapsdev-api.arcgis.com/arcgis/rest/services/styles/ArcGIS:HumanGeographyDark",
            "title": "HumanGeographyDark",
        }
    ],
}



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# --------------------------------------------------------
# References:
# timm: https://github.com/rwightman/pytorch-image-models/tree/master/timm
# DeiT: https://github.com/facebookresearch/deit
# --------------------------------------------------------

import numpy as np
import torch
import torch.nn as nn
from mmcv.runner import load_checkpoint
from mmseg.models.builder import BACKBONES, NECKS
from timm.models.layers import to_2tuple
from timm.models.vision_transformer import Block
from mmseg.ops import resize
from mmseg.models import builder
from mmseg.models.builder import SEGMENTORS
from mmseg.models.segmentors.encoder_decoder import EncoderDecoder


def _convTranspose2dOutput(
    input_size: int,
    stride: int,
    padding: int,
    dilation: int,
    kernel_size: int,
    output_padding: int,
):
    """
    Calculate the output size of a ConvTranspose2d.
    Taken from: https://pytorch.org/docs/stable/generated/torch.nn.ConvTranspose2d.html
    """
    return (
        (input_size - 1) * stride
        - 2 * padding
        + dilation * (kernel_size - 1)
        + output_padding
        + 1
    )


def get_1d_sincos_pos_embed_from_grid(embed_dim: int, pos: torch.Tensor):
    """
    embed_dim: output dimension for each position
    pos: a list of positions to be encoded: size (M,)
    out: (M, D)
    """
    assert embed_dim % 2 == 0
    omega = np.arange(embed_dim // 2, dtype=np.float32)
    omega /= embed_dim / 2.0
    omega = 1.0 / 10000**omega  # (D/2,)

    pos = pos.reshape(-1)  # (M,)
    out = np.einsum("m,d->md", pos, omega)  # (M, D/2), outer product

    emb_sin = np.sin(out)  # (M, D/2)
    emb_cos = np.cos(out)  # (M, D/2)

    emb = np.concatenate([emb_sin, emb_cos], axis=1)  # (M, D)
    return emb


def get_3d_sincos_pos_embed(embed_dim: int, grid_size: tuple, cls_token: bool = False):
    # Copyright (c) Meta Platforms, Inc. and affiliates.
    # All rights reserved.

    # This source code is licensed under the license found in the
    # LICENSE file in the root directory of this source tree.
    # --------------------------------------------------------
    # Position embedding utils
    # --------------------------------------------------------
    """
    grid_size: 3d tuple of grid size: t, h, w
    return:
    pos_embed: L, D
    """

    assert embed_dim % 16 == 0

    t_size, h_size, w_size = grid_size

    w_embed_dim = embed_dim // 16 * 6
    h_embed_dim = embed_dim // 16 * 6
    t_embed_dim = embed_dim // 16 * 4

    w_pos_embed = get_1d_sincos_pos_embed_from_grid(w_embed_dim, np.arange(w_size))
    h_pos_embed = get_1d_sincos_pos_embed_from_grid(h_embed_dim, np.arange(h_size))
    t_pos_embed = get_1d_sincos_pos_embed_from_grid(t_embed_dim, np.arange(t_size))

    w_pos_embed = np.tile(w_pos_embed, (t_size * h_size, 1))
    h_pos_embed = np.tile(np.repeat(h_pos_embed, w_size, axis=0), (t_size, 1))
    t_pos_embed = np.repeat(t_pos_embed, h_size * w_size, axis=0)

    pos_embed = np.concatenate((w_pos_embed, h_pos_embed, t_pos_embed), axis=1)

    if cls_token:
        pos_embed = np.concatenate([np.zeros([1, embed_dim]), pos_embed], axis=0)
    return pos_embed


class PatchEmbed(nn.Module):
    """Frames of 2D Images to Patch Embedding
    The 3D version of timm.models.vision_transformer.PatchEmbed
    """

    def __init__(
        self,
        img_size: int = 224,
        patch_size: int = 16,
        num_frames: int = 3,
        tubelet_size: int = 1,
        in_chans: int = 3,
        embed_dim: int = 768,
        norm_layer: nn.Module = None,
        flatten: bool = True,
        bias: bool = True,
    ):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_frames = num_frames
        self.in_chans = in_chans
        self.tubelet_size = tubelet_size
        self.grid_size = (
            num_frames // tubelet_size,
            img_size[0] // patch_size[0],
            img_size[1] // patch_size[1],
        )
        self.num_patches = self.grid_size[0] * self.grid_size[1] * self.grid_size[2]
        self.flatten = flatten

        self.proj = nn.Conv3d(
            in_chans,
            embed_dim,
            kernel_size=(tubelet_size, patch_size[0], patch_size[1]),
            stride=(tubelet_size, patch_size[0], patch_size[1]),
            bias=bias,
        )
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

    def forward(self, x):
        if self.num_frames > 1:
            if x.shape[1] == 6:
                x = (
                    x.unsqueeze(1)
                    .repeat(1, self.num_frames, 1, 1, 1)
                    .view(
                        x.shape[0], x.shape[1], self.num_frames, x.shape[2], x.shape[3]
                    )
                )
            else:
                x = torch.reshape(
                    x,
                    (
                        x.shape[0],
                        self.in_chans,
                        self.num_frames,
                        x.shape[2],
                        x.shape[3],
                    ),
                )
        else:
            x = x[:, :, None, :, :]
        B, C, T, H, W = x.shape
        assert (
            H == self.img_size[0]
        ), f"Input image height ({H}) doesn't match model ({self.img_size[0]})."
        assert (
            W == self.img_size[1]
        ), f"Input image width ({W}) doesn't match model ({self.img_size[1]})."
        x = self.proj(x)
        Hp, Wp = x.shape[3], x.shape[4]
        if self.flatten:
            x = x.flatten(2).transpose(1, 2)  # B,C,T,H,W -> B,C,L -> B,L,C
        x = self.norm(x)
        return x, Hp, Wp


class Norm2d(nn.Module):
    def __init__(self, embed_dim: int):
        super().__init__()
        self.ln = nn.LayerNorm(embed_dim, eps=1e-6)

    def forward(self, x):
        x = x.permute(0, 2, 3, 1)
        x = self.ln(x)
        x = x.permute(0, 3, 1, 2).contiguous()
        return x


@NECKS.register_module()
class GeospatialNeck(nn.Module):
    """
    Neck that transforms the token-based output of transformer into a single embedding suitable for processing with standard layers.
    Performs 4 ConvTranspose2d operations on the rearranged input with kernel_size=2 and stride=2
    """

    def __init__(
        self,
        embed_dim: int,
        first_conv_channels: int,
        Hp: int = 14,
        Wp: int = 14,
        channel_reduction_factor: int = 2,
        num_convs: int = 4,
        num_convs_per_upscale: int = 1,
        dropout: bool = False,
        drop_cls_token: bool = True,
    ):
        """

        Args:
            embed_dim (int): Input embedding dimension
            first_conv_channel (int): Number of channels for first dimension
            Hp (int, optional): Height (in patches) of embedding to be upscaled. Defaults to 14.
            Wp (int, optional): Width (in patches) of embedding to be upscaled. Defaults to 14.
            channel_reduction_factor (int): Factor that each convolutional block reduces number of channels by.
            num_convs (int): Number of convolutional upscaling blocks. Each upscales 2x.
            drop_cls_token (bool, optional): Whether there is a cls_token, which should be dropped. This assumes the cls token is the first token. Defaults to True.
        """
        super().__init__()
        self.drop_cls_token = drop_cls_token
        self.Hp = Hp
        self.Wp = Wp
        self.H_out = Hp
        self.W_out = Wp
        self.dropout = dropout

        conv_kernel_size = 3
        conv_padding = 1

        kernel_size = 2
        stride = 2
        dilation = 1
        padding = 0
        output_padding = 0

        self.embed_dim = embed_dim
        self.channels = [
            first_conv_channels // (channel_reduction_factor**i)
            for i in range(num_convs)
        ]
        self.channels = [embed_dim] + self.channels

        for _ in range(len(self.channels) - 1):
            self.H_out = _convTranspose2dOutput(
                self.H_out, stride, padding, dilation, kernel_size, output_padding
            )
            self.W_out = _convTranspose2dOutput(
                self.W_out, stride, padding, dilation, kernel_size, output_padding
            )

        def _build_upscale_block(channels_in, channels_out):
            layers = []
            layers.append(
                nn.ConvTranspose2d(
                    channels_in,
                    channels_out,
                    kernel_size=kernel_size,
                    stride=stride,
                    dilation=dilation,
                    padding=padding,
                    output_padding=output_padding,
                )
            )

            layers += [
                nn.Sequential(
                    nn.Conv2d(
                        channels_out,
                        channels_out,
                        kernel_size=conv_kernel_size,
                        padding=conv_padding,
                    ),
                    nn.BatchNorm2d(channels_out),
                    nn.Dropout() if self.dropout else nn.Identity(),
                    nn.ReLU(),
                )
                for _ in range(num_convs_per_upscale)
            ]

            return nn.Sequential(*layers)

        self.layers = nn.ModuleList(
            [
                _build_upscale_block(self.channels[i], self.channels[i + 1])
                for i in range(len(self.channels) - 1)
            ]
        )

    def forward(self, x):
        x = x[0]
        if self.drop_cls_token:
            x = x[:, 1:, :]
        x = x.permute(0, 2, 1).reshape(x.shape[0], -1, self.Hp, self.Wp)

        for layer in self.layers:
            x = layer(x)

        x = x.reshape((x.shape[0], self.channels[-1], self.H_out, self.W_out))

        out = tuple([x])

        return out


@NECKS.register_module()
class ConvTransformerTokensToEmbeddingNeck(nn.Module):
    """
    Neck that transforms the token-based output of transformer into a single embedding suitable for processing with standard layers.
    Performs 4 ConvTranspose2d operations on the rearranged input with kernel_size=2 and stride=2
    """

    def __init__(
        self,
        embed_dim: int,
        output_embed_dim: int,
        # num_frames: int = 1,
        Hp: int = 14,
        Wp: int = 14,
        drop_cls_token: bool = True,
    ):
        """

        Args:
            embed_dim (int): Input embedding dimension
            output_embed_dim (int): Output embedding dimension
            Hp (int, optional): Height (in patches) of embedding to be upscaled. Defaults to 14.
            Wp (int, optional): Width (in patches) of embedding to be upscaled. Defaults to 14.
            drop_cls_token (bool, optional): Whether there is a cls_token, which should be dropped. This assumes the cls token is the first token. Defaults to True.
        """
        super().__init__()
        self.drop_cls_token = drop_cls_token
        self.Hp = Hp
        self.Wp = Wp
        self.H_out = Hp
        self.W_out = Wp
        # self.num_frames = num_frames

        kernel_size = 2
        stride = 2
        dilation = 1
        padding = 0
        output_padding = 0
        for _ in range(4):
            self.H_out = _convTranspose2dOutput(
                self.H_out, stride, padding, dilation, kernel_size, output_padding
            )
            self.W_out = _convTranspose2dOutput(
                self.W_out, stride, padding, dilation, kernel_size, output_padding
            )

        self.embed_dim = embed_dim
        self.output_embed_dim = output_embed_dim
        self.fpn1 = nn.Sequential(
            nn.ConvTranspose2d(
                self.embed_dim,
                self.output_embed_dim,
                kernel_size=kernel_size,
                stride=stride,
                dilation=dilation,
                padding=padding,
                output_padding=output_padding,
            ),
            Norm2d(self.output_embed_dim),
            nn.GELU(),
            nn.ConvTranspose2d(
                self.output_embed_dim,
                self.output_embed_dim,
                kernel_size=kernel_size,
                stride=stride,
                dilation=dilation,
                padding=padding,
                output_padding=output_padding,
            ),
        )
        self.fpn2 = nn.Sequential(
            nn.ConvTranspose2d(
                self.output_embed_dim,
                self.output_embed_dim,
                kernel_size=kernel_size,
                stride=stride,
                dilation=dilation,
                padding=padding,
                output_padding=output_padding,
            ),
            Norm2d(self.output_embed_dim),
            nn.GELU(),
            nn.ConvTranspose2d(
                self.output_embed_dim,
                self.output_embed_dim,
                kernel_size=kernel_size,
                stride=stride,
                dilation=dilation,
                padding=padding,
                output_padding=output_padding,
            ),
        )

    def forward(self, x):
        x = x[0]
        if self.drop_cls_token:
            x = x[:, 1:, :]
        x = x.permute(0, 2, 1).reshape(x.shape[0], -1, self.Hp, self.Wp)

        x = self.fpn1(x)
        x = self.fpn2(x)

        x = x.reshape((-1, self.output_embed_dim, self.H_out, self.W_out))

        out = tuple([x])

        return out


@BACKBONES.register_module()
class TemporalViTEncoder(nn.Module):
    """Encoder from an ViT with capability to take in temporal input.

    This class defines an encoder taken from a ViT architecture.
    """

    def __init__(
        self,
        img_size: int = 224,
        patch_size: int = 16,
        num_frames: int = 1,
        tubelet_size: int = 1,
        in_chans: int = 3,
        embed_dim: int = 1024,
        depth: int = 24,
        num_heads: int = 16,
        mlp_ratio: float = 4.0,
        norm_layer: nn.Module = nn.LayerNorm,
        norm_pix_loss: bool = False,
        pretrained: str = None,
    ):
        """

        Args:
            img_size (int, optional): Input image size. Defaults to 224.
            patch_size (int, optional): Patch size to be used by the transformer. Defaults to 16.
            num_frames (int, optional): Number of frames (temporal dimension) to be input to the encoder. Defaults to 1.
            tubelet_size (int, optional): Tubelet size used in patch embedding. Defaults to 1.
            in_chans (int, optional): Number of input channels. Defaults to 3.
            embed_dim (int, optional): Embedding dimension. Defaults to 1024.
            depth (int, optional): Encoder depth. Defaults to 24.
            num_heads (int, optional): Number of heads used in the encoder blocks. Defaults to 16.
            mlp_ratio (float, optional): Ratio to be used for the size of the MLP in encoder blocks. Defaults to 4.0.
            norm_layer (nn.Module, optional): Norm layer to be used. Defaults to nn.LayerNorm.
            norm_pix_loss (bool, optional): Whether to use Norm Pix Loss. Defaults to False.
            pretrained (str, optional): Path to pretrained encoder weights. Defaults to None.
        """
        super().__init__()

        # --------------------------------------------------------------------------
        # MAE encoder specifics
        self.embed_dim = embed_dim
        self.patch_embed = PatchEmbed(
            img_size, patch_size, num_frames, tubelet_size, in_chans, embed_dim
        )
        num_patches = self.patch_embed.num_patches
        self.num_frames = num_frames

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(
            torch.zeros(1, num_patches + 1, embed_dim), requires_grad=False
        )  # fixed sin-cos embedding

        self.blocks = nn.ModuleList(
            [
                Block(
                    embed_dim,
                    num_heads,
                    mlp_ratio,
                    qkv_bias=True,
                    norm_layer=norm_layer,
                )
                for _ in range(depth)
            ]
        )
        self.norm = norm_layer(embed_dim)

        self.norm_pix_loss = norm_pix_loss
        self.pretrained = pretrained

        self.initialize_weights()

    def initialize_weights(self):
        # initialization
        # initialize (and freeze) pos_embed by sin-cos embedding
        pos_embed = get_3d_sincos_pos_embed(
            self.pos_embed.shape[-1], self.patch_embed.grid_size, cls_token=True
        )
        self.pos_embed.data.copy_(torch.from_numpy(pos_embed).float().unsqueeze(0))

        # initialize patch_embed like nn.Linear (instead of nn.Conv2d)
        w = self.patch_embed.proj.weight.data
        torch.nn.init.xavier_uniform_(w.view([w.shape[0], -1]))

        if isinstance(self.pretrained, str):
            self.apply(self._init_weights)
            # print(f"load from {self.pretrained}")
            load_checkpoint(self, self.pretrained, strict=False)
        elif self.pretrained is None:
            # # initialize nn.Linear and nn.LayerNorm
            self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward(self, x):
        # embed patches
        x, _, _ = self.patch_embed(x)

        # add pos embed w/o cls token
        x = x + self.pos_embed[:, 1:, :]

        # append cls token
        cls_token = self.cls_token + self.pos_embed[:, :1, :]
        cls_tokens = cls_token.expand(x.shape[0], -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)

        # apply Transformer blocks
        for blk in self.blocks:
            x = blk(x)

        x = self.norm(x)

        return tuple([x])


@SEGMENTORS.register_module()
class TemporalEncoderDecoder(EncoderDecoder):
    """Encoder Decoder segmentors.

    EncoderDecoder typically consists of backbone, neck, decode_head, auxiliary_head.
    Note that auxiliary_head is only used for deep supervision during training,
    which could be dumped during inference.

    The backbone should return plain embeddings.
    The neck can process these to make them suitable for the chosen heads.
    The heads perform the final processing that will return the output.
    """

    def __init__(
        self,
        backbone,
        decode_head,
        neck=None,
        auxiliary_head=None,
        train_cfg=None,
        test_cfg=None,
        pretrained=None,
        init_cfg=None,
        frozen_backbone=False,
    ):
        super(EncoderDecoder, self).__init__(init_cfg)
        if pretrained is not None:
            assert (
                backbone.get("pretrained") is None
            ), "both backbone and segmentor set pretrained weight"
            backbone.pretrained = pretrained
        self.backbone = builder.build_backbone(backbone)

        if frozen_backbone:
            for param in self.backbone.parameters():
                param.requires_grad = False

        if neck is not None:
            self.neck = builder.build_neck(neck)
        self._init_decode_head(decode_head)
        self._init_auxiliary_head(auxiliary_head)

        self.train_cfg = train_cfg
        self.test_cfg = test_cfg
        assert self.with_decode_head

    def encode_decode(self, img, img_metas):
        """Encode images with backbone and decode into a semantic segmentation
        map of the same size as input."""
        x = self.extract_feat(img)
        out = self._decode_head_forward_test(x, img_metas)

        #### size calculated over last two dimensions ###
        size = img.shape[-2:]

        out = resize(
            input=out, size=size, mode="bilinear", align_corners=self.align_corners
        )

        return out



import traceback
from .._data import _raise_fastai_import_error
from ._inference_only_models import InferenceOnlyModel

HAS_TRANSFORMER = True

try:
    import torch
    from transformers import pipeline, logging
    from fastprogress.fastprogress import progress_bar

    try:
        # For version 3.3.0
        from transformers.modeling_auto import MODEL_FOR_MASKED_LM_MAPPING
    except ModuleNotFoundError as e:
        # For version 4.5.1
        from transformers.models.auto.modeling_auto import MODEL_FOR_MASKED_LM_MAPPING
    EXPECTED_MODEL_TYPES = [
        x.__name__.replace("Config", "") for x in MODEL_FOR_MASKED_LM_MAPPING.keys()
    ]
except Exception as e:
    transformer_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_TRANSFORMER = False
    EXPECTED_MODEL_TYPES = []


class FillMask(InferenceOnlyModel):
    """
    Creates a :class:`~arcgis.learn.text.FillMask` Object.
    Based on the Hugging Face transformers library

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    backbone                Optional string. Specify the HuggingFace
                            transformer model name which will be used to
                            generate the suggestion token.

                            To learn more about the available models for
                            fill-mask task, kindly visit:-
                            https://huggingface.co/models?pipeline_tag=fill-mask
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    pretrained_path         Option str. Path to a directory, where pretrained
                            model files are saved.
                            If pretrained_path is provided, the model is
                            loaded from that path on the local disk.
    ---------------------   -------------------------------------------
    working_dir             Option str. Path to a directory on local filesystem.
                            If directory is not present, it will be created.
                            This directory is used as the location to save the
                            model.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.text.FillMask` Object
    """

    #: supported transformer architectures
    supported_backbones = EXPECTED_MODEL_TYPES

    def __init__(self, backbone=None, **kwargs):
        if not HAS_TRANSFORMER:
            _raise_fastai_import_error(import_exception=transformer_exception)
        super().__init__(backbone=backbone, task="fill-mask", **kwargs)

    def _load_model(self):
        try:
            # For version 3.3.0
            if self._pretrained_path:
                self.model = pipeline(
                    self._task,
                    model=self._pretrained_path,
                    device=self._device,
                    topk=10,
                )
            else:
                self.model = pipeline(
                    self._task, model=self._backbone, device=self._device, topk=10
                )
        except TypeError as e:
            # For version 4.5.1
            if self._pretrained_path:
                self.model = pipeline(
                    self._task,
                    model=self._pretrained_path,
                    device=self._device,
                    top_k=10,
                )
            else:
                self.model = pipeline(
                    self._task, model=self._backbone, device=self._device, top_k=10
                )
        except Exception as e:
            error_message = (
                f"`{self._backbone}` is not valid backbone name for {self._task} task.\n"
                f"For selecting backbone name for {self._task} task, kindly visit:- "
                f"https://huggingface.co/models?pipeline_tag={self._task} "
            )
            raise Exception(error_message)

    def predict_token(self, text_or_list, num_suggestions=5, show_progress=True):
        """
        Summarize the given text or list of text

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required string or list. A text/sentence
                                or a list of texts/sentences for which on wishes
                                to generate the recommendations for masked-token.
        ---------------------   -------------------------------------------
        num_suggestions         Optional Integer. The number of suggestions to
                                return. The maximum number of suggestion that
                                can be generated for a `missing-token` is 10.
        ---------------------   -------------------------------------------
        show_progress           optional Bool. If set to True, will display a
                                progress bar depicting the items processed so far.
        =====================   ===========================================

        :return: A list or a list of list of :obj:`dict`: Each result comes as list of dictionaries with the following keys:

            - **sequence** (:obj:`str`)  -- The corresponding input with the mask token prediction.
            - **score** (:obj:`float`)   -- The corresponding probability.
            - **token_str** (:obj:`str`) -- The predicted token (to replace the masked one).
        """
        results = []
        if not isinstance(text_or_list, (list, tuple)):
            text_or_list = [text_or_list]
        self._do_sanity(text_or_list)

        for i in progress_bar(range(len(text_or_list)), display=show_progress):
            text = text_or_list[i].replace("__", self.model.tokenizer.mask_token)
            result = self._process_result(self.model(text)[:num_suggestions])
            if num_suggestions == 1:
                result = result[0]
            results.append(result)
        return results

    @staticmethod
    def _do_sanity(text_list):
        for text in text_list:
            if "__" not in text:
                error_message = (
                    f"Text - `{text}` is mising `__` token. Use the `__` token to "
                    "specify where you want to generate the suggestions in the text."
                )
                raise Exception(error_message)

    def _process_result(self, result_list):
        for item in result_list:
            _ = item.pop("token", -1)
            token = item["token_str"]
            if isinstance(token, (str, bytes)):
                token = [token]
            sequence = item["sequence"]
            item["token_str"] = self.model.tokenizer.convert_tokens_to_string(
                token
            ).strip()
            item["sequence"] = self.model.tokenizer.decode(
                self.model.tokenizer.encode(sequence, add_special_tokens=False),
                skip_special_tokens=True,
            )

        return result_list



import uuid
import arcgis
from .._utils._basewidget import _BaseWidget
from .._utils._basewidget import NoDataProperties


class EmbeddedContent(_BaseWidget):
    """
    Creates a dashboard Embedded Content Widget.

    =========================   ===========================================
    **Parameter**                **Description**
    -------------------------   -------------------------------------------
    url                         Required string. Url of the embedded content
                                or field name if item is not None.
    -------------------------   -------------------------------------------
    name                        Optional string. Name of the widget.
    -------------------------   -------------------------------------------
    title                       Optional string. Title of the widget.
    -------------------------   -------------------------------------------
    description                 Optional string. Description of the widget.
    -------------------------   -------------------------------------------
    content_type                Optional string. Type of the content.
                                Choose from "document", "image", "video".
    -------------------------   -------------------------------------------
    refresh_interval            Optional integer. Interval to refresh in
                                minutes. It is only applicable for
                                content_type = 'image'
    -------------------------   -------------------------------------------
    item                        Optional Portal :class:`~arcgis.gis.Item` . To show content from
                                portal.
    -------------------------   -------------------------------------------
    layer                       Optional integer. Layer number when item is
                                a mapwidget.
    =========================   ===========================================
    """

    def __init__(
        self,
        url,
        name="EmbeddedContent",
        title="",
        description="",
        content_type="document",
        refresh_interval=0,
        item=None,
        layer=0,
    ):
        super().__init__(name, title, description)

        data_type = "features"

        if item is None:
            data_type = "static"
            self._url = url
        else:
            self._url = "{" + url + "}"

        self.item = item
        self.layer = layer
        self.type = "embeddedContentWidget"

        self._data_type = data_type

        if content_type not in ["document", "image", "video"]:
            raise Exception("Invalid content type.")

        self._content_type = content_type
        self._max_features_displayed = 50
        self._refresh_interval = refresh_interval

        self._no_data = NoDataProperties._nodata_init()

    @classmethod
    def _from_json(cls, widget_json):
        from arcgis.apps.dashboard import EmbeddedContent

        gis = arcgis.env.active_gis
        itemid = widget_json["datasets"]["datasource"]["itemid"]
        name = widget_json["name"]
        item = gis.content.get(itemid)
        title = widget_json["caption"]
        description = widget_json["description"]
        emb = EmbeddedContent(name, title, description, item)

        return emb

    @property
    def url(self):
        """
        :return: Url of the embedded content.
        """
        return self._url

    @url.setter
    def url(self, value):
        """
        Set Url for the embedded content.
        """
        self._url = value

    @property
    def content_type(self):
        """
        :return: Content type of the embedded content.
        """
        return self._content_type

    @content_type.setter
    def content_type(self, value):
        """
        Set Content type for the embedded content.
        """
        if value in ["document", "image", "video"]:
            self._content_type = value
        else:
            raise Exception("Invalid content type")

    @property
    def max_features(self):
        """
        :return: Max Features to display for feature data.
        """
        return self._max_features_displayed

    @max_features.setter
    def max_features(self, value):
        """
        Set max features to display for data from features.
        """
        self._max_features_displayed = value

    @property
    def refresh_interval(self):
        """
        :return: Refresh interval for image content.
        """
        return self._refresh_interval

    @refresh_interval.setter
    def refresh_interval(self, value):
        """
        Set refresh interval for image content.
        """
        self._refresh_interval = value

    @property
    def no_data(self):
        """
        :return: NoDataProperties Object
        """
        return self._no_data

    def _convert_to_json(self):
        json_data = {
            "type": "embeddedContentWidget",
            "url": self._url,
            "contentType": self._content_type,
            "imageRefreshInterval": self._refresh_interval,
            "videoSettings": {
                "controls": True,
                "autoplay": False,
                "loop": False,
                "muted": False,
                "controlsList": "nodownload",
            },
            "datasets": [],
            "id": self._id,
            "name": self.name,
            "caption": self._title,
            "description": self.description,
            "showLastUpdate": True,
            "noDataVerticalAlignment": self._no_data._alignment,
            "showCaptionWhenNoData": self._no_data._show_title,
            "showDescriptionWhenNoData": self._no_data._show_description,
        }

        if self._no_data._text:
            json_data["noDataText"] = self._no_data._text

        if self._background_color:
            json_data["backgroundColor"] = self._background_color

        if self._text_color:
            json_data["textColor"] = self._text_color

        if self.item:
            json_data["datasets"] = [
                {
                    "type": "serviceDataset",
                    "dataSource": {
                        "type": "featureServiceDataSource",
                        "itemId": self.item.itemid,
                        "layerId": self.layer,
                        "table": True,
                    },
                    "outFields": ["*"],
                    "groupByFields": [],
                    "orderByFields": [],
                    "statisticDefinitions": [],
                    "maxFeatures": self._max_features_displayed,
                    "querySpatialRelationship": "esriSpatialRelIntersects",
                    "returnGeometry": False,
                    "clientSideStatistics": False,
                    "name": "main",
                }
            ]

        return json_data
